# -*- coding: utf-8 -*-
"""Base Isotope class."""
from os import path
from re import match

from monty.serialization import loadfn
from pydantic import BaseModel
from pydantic import validator

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"

MODULE_DIR = path.dirname(path.abspath(__file__))
ISOTOPE_DATA = loadfn(path.join(MODULE_DIR, "isotope_data.json"))


class Isotope(BaseModel):
    """The Isotope class.

    Attributes
    ----------

    symbol: str (required)
        The isotope symbol given as the atomic number followed by the atomic symbol.

    Example
    -------

    >>> # 13C isotope information
    >>> carbon = Isotope(symbol='13C')
    >>> carbon.spin
    0.5
    >>> carbon.natural_abundance # in %
    1.11
    >>> carbon.gyromagnetic_ratio # in MHz/T
    10.7084
    >>> carbon.atomic_number
    6
    >>> carbon.quadrupole_moment # in eB
    0.0

    """

    symbol: str

    class Config:
        validate_assignment = True

    @validator("symbol", always=True)
    def get_isotope(cls, v, *, values, **kwargs):
        return format_isotope_string(v)

    def to_dict_with_units(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit."""
        return self.symbol

    @property
    def spin(self):
        """Spin quantum number, I, of the isotope."""
        isotope_data = get_isotope_data(self.symbol)
        return isotope_data["spin"] / 2.0

    @property
    def natural_abundance(self):
        """Natural abundance of the isotope in units of %."""
        isotope_data = get_isotope_data(self.symbol)
        return isotope_data["natural_abundance"]

    @property
    def gyromagnetic_ratio(self):
        """Reduced gyromagnetic ratio of the nucleus given in units of MHz/T."""
        isotope_data = get_isotope_data(self.symbol)
        return isotope_data["gyromagnetic_ratio"]

    @property
    def quadrupole_moment(self):
        """Quadrupole moment of the nucleus given in units of eB (electron-barn)."""
        isotope_data = get_isotope_data(self.symbol)
        return isotope_data["quadrupole_moment"]

    @property
    def atomic_number(self):
        """Atomic number of the isotope."""
        isotope_data = get_isotope_data(self.symbol)
        return isotope_data["atomic_number"]


def format_isotope_string(isotope_string: str) -> str:
    """Format the isotope string to {A}{symbol}, where A is the isotope number."""
    result = match(r"(\d+)\s*(\w+)", isotope_string)

    if result is None:
        raise Exception(f"Could not parse isotope string {isotope_string}")
    isotope = result.group(2)
    A = result.group(1)

    formatted_string = f"{A}{isotope}"
    if formatted_string not in ISOTOPE_DATA:
        raise Exception(f"Could not parse isotope string {isotope_string}")

    return formatted_string


def get_isotope_data(isotope_string: str) -> dict:
    """Get the isotope's intrinsinc properties from a JSON data file."""
    formatted_isotope_string = format_isotope_string(isotope_string)
    isotope_dict = dict(ISOTOPE_DATA[formatted_isotope_string])
    isotope_dict.update({"isotope": formatted_isotope_string})
    return isotope_dict



# -*- coding: utf-8 -*-
"""Base Simulator class."""
import json
from copy import deepcopy
from typing import List

import csdmpy as cp
import numpy as np
from mrsimulator import __version__
from mrsimulator import SpinSystem
from mrsimulator.base_model import one_d_spectrum
from mrsimulator.method import Method
from mrsimulator.utils import _reduce_dict
from mrsimulator.utils.importer import import_json
from pydantic import BaseModel

from .config import ConfigSimulator

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class Simulator(BaseModel):
    """
    The simulator class.

    Attributes
    ----------

    spin_systems: A list of :ref:`spin_sys_api` or equivalent dict objects (optional).
        The value is a list of NMR spin systems present within the sample, where each
        spin system is an isolated system. The default value is an empty list.

        Example
        -------

        >>> sim = Simulator()
        >>> sim.spin_systems = [
        ...     SpinSystem(sites=[Site(isotope='17O')], abundance=0.015),
        ...     SpinSystem(sites=[Site(isotope='1H')], abundance=1),
        ... ]
        >>> # or equivalently
        >>> sim.spin_systems = [
        ...     {'sites': [{'isotope': '17O'}], 'abundance': 0.015},
        ...     {'sites': [{'isotope': '1H'}], 'abundance': 1},
        ... ]

    methods: A list of :ref:`method_api` (optional).
        The value is a list of NMR methods. The default value is an empty list.

        Example
        -------

        >>> from mrsimulator.methods import BlochDecaySpectrum
        >>> from mrsimulator.methods import BlochDecayCentralTransitionSpectrum
        >>> sim.methods = [
        ...     BlochDecaySpectrum(channels=['17O'], spectral_width=50000),
        ...     BlochDecayCentralTransitionSpectrum(channels=['17O'])
        ... ]

    config: :ref:`config_api` object or equivalent dict object (optional).
        The :ref:`config_api` object is used to configure the simulation. The valid
        attributes of the ConfigSimulator object are

        - ``number_of_sidebands``,
        - ``integration_density``,
        - ``integration_volume``, and
        - ``decompose_spectrum``

        Example
        -------

        >>> from mrsimulator.simulator.config import ConfigSimulator
        >>> sim.config = ConfigSimulator(
        ...     number_of_sidebands=32,
        ...     integration_density=64,
        ...     integration_volume='hemisphere',
        ...     decompose_spectrum='spin_system',
        ... )
        >>> # or equivalently
        >>> sim.config = {
        ...     'number_of_sidebands': 32,
        ...     'integration_density': 64,
        ...     'integration_volume': 'hemisphere',
        ...     'decompose_spectrum': 'spin_system',
        ... }

        See :ref:`config_simulator` for details.

    name: str (optional).
        The value is the name or id of the simulation or sample. The default value is
        None.

        Example
        -------

        >>> sim.name = '1H-17O'
        >>> sim.name
        '1H-17O'

    label: str (optional).
        The value is a label for the simulation or sample. The default value is None.

        Example
        -------

        >>> sim.label = 'Test simulator'
        >>> sim.label
        'Test simulator'

    description: str (optional).
        The value is a description of the simulation or sample. The default value is
        None.

        Example
        -------

        >>> sim.description = 'Simulation for sample 1'
        >>> sim.description
        'Simulation for sample 1'

    """

    name: str = None
    label: str = None
    description: str = None
    spin_systems: List[SpinSystem] = []
    methods: List[Method] = []
    config: ConfigSimulator = ConfigSimulator()
    indexes = []

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse the physical quantity from a dictionary representation of the Simulator
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A required python dict object.

        Returns:
            A :ref:`simulator_api` object.

        Example
        -------

        >>> sim_py_dict = {
        ...     'config': {
        ...         'decompose_spectrum': 'none',
        ...         'integration_density': 70,
        ...         'integration_volume': 'octant',
        ...         'number_of_sidebands': 64
        ...     },
        ...     'spin_systems': [
        ...         {
        ...             'abundance': '100 %',
        ...             'sites': [{
        ...                 'isotope': '13C',
        ...                 'isotropic_chemical_shift': '20.0 ppm',
        ...                 'shielding_symmetric': {'eta': 0.5, 'zeta': '10.0 ppm'}
        ...             }]
        ...         },
        ...         {
        ...             'abundance': '100 %',
        ...             'sites': [{
        ...                 'isotope': '1H',
        ...                     'isotropic_chemical_shift': '-4.0 ppm',
        ...                     'shielding_symmetric': {'eta': 0.1, 'zeta': '2.1 ppm'}
        ...             }]
        ...         },
        ...         {
        ...             'abundance': '100 %',
        ...             'sites': [{
        ...                 'isotope': '27Al',
        ...                 'isotropic_chemical_shift': '120.0 ppm',
        ...                 'shielding_symmetric': {'eta': 0.1, 'zeta': '2.1 ppm'}
        ...             }]
        ...         }
        ...     ]
        ... }
        >>> sim = Simulator.parse_dict_with_units(sim_py_dict)
        >>> len(sim.spin_systems)
        3
        """
        py_copy_dict = deepcopy(py_dict)

        if "spin_systems" in py_copy_dict:
            spin_sys = py_copy_dict["spin_systems"]
            spin_sys = [SpinSystem.parse_dict_with_units(obj) for obj in spin_sys]
            py_copy_dict["spin_systems"] = spin_sys

        if "methods" in py_copy_dict:
            methods = py_copy_dict["methods"]
            methods = [Method.parse_dict_with_units(obj) for obj in methods]
            py_copy_dict["methods"] = methods

        return Simulator(**py_copy_dict)

    def get_isotopes(self, spin_I=None) -> set:
        """
        Set of unique isotopes from the sites within the list of the spin systems
        corresponding to spin quantum number `I`. If `I` is None, a set of all unique
        isotopes is returned instead.

        Args:
            float spin_I: An optional spin quantum number. The valid input are the
                multiples of 0.5.

        Returns:
            A Set.

        Example
        -------

        >>> sim.get_isotopes() # doctest:+SKIP
        {'1H', '27Al', '13C'}
        >>> sim.get_isotopes(spin_I=0.5) # doctest:+SKIP
        {'1H', '13C'}
        >>> sim.get_isotopes(spin_I=1.5)
        set()
        >>> sim.get_isotopes(spin_I=2.5)
        {'27Al'}
        """
        st = set()
        for spin_system in self.spin_systems:
            st.update(spin_system.get_isotopes(spin_I))
        return st

    def to_dict_with_units(
        self, include_methods: bool = False, include_version: bool = False
    ):
        """
        Serialize the Simulator object to a JSON compliant python dictionary object
        where physical quantities are represented as string with a value and a unit.

        Args:
            bool include_methods: If True, the output dictionary will include the
                serialized method objects. The default value is False.
            bool include_version: If True, add a version key-value pair to the
                serialized output dictionary. The default is False.

        Returns:
            A Dict object.

        Example
        -------

        >>> pprint(sim.to_dict_with_units())
        {'config': {'decompose_spectrum': 'none',
                    'integration_density': 70,
                    'integration_volume': 'octant',
                    'number_of_sidebands': 64},
         'indexes': [],
         'spin_systems': [{'abundance': '100 %',
                           'sites': [{'isotope': '13C',
                                      'isotropic_chemical_shift': '20.0 ppm',
                                      'shielding_symmetric': {'eta': 0.5,
                                                              'zeta': '10.0 ppm'}}]},
                          {'abundance': '100 %',
                           'sites': [{'isotope': '1H',
                                      'isotropic_chemical_shift': '-4.0 ppm',
                                      'shielding_symmetric': {'eta': 0.1,
                                                              'zeta': '2.1 ppm'}}]},
                          {'abundance': '100 %',
                           'sites': [{'isotope': '27Al',
                                      'isotropic_chemical_shift': '120.0 ppm',
                                      'shielding_symmetric': {'eta': 0.1,
                                                              'zeta': '2.1 ppm'}}]}]}
        """
        sim = {}

        if self.name is not None:
            sim["name"] = self.name

        if self.description is not None:
            sim["description"] = self.description

        if self.label is not None:
            sim["label"] = self.label

        sim["spin_systems"] = [_.to_dict_with_units() for _ in self.spin_systems]

        if include_methods:
            method = [_.to_dict_with_units() for _ in self.methods]
            if len(method) != 0:
                sim["methods"] = method

        sim["config"] = self.config.dict()
        sim["indexes"] = self.indexes
        if include_version:
            sim["version"] = __version__
        return sim

    def reduced_dict(self, exclude=["property_units"]) -> dict:
        """Returns a reduced dictionary representation of the class object by removing
        all key-value pair corresponding to keys listed in the `exclude` argument, and
        keys with value as None.

        Args:
            list exclude: A list of keys to exclude from the dictionary.
        Return: A dict.
         """
        return _reduce_dict(self.dict(), exclude)

    def load_spin_systems(self, filename: str):
        """
        Load a list of spin systems from the given JSON serialized file.

        See an
        `example <https://raw.githubusercontent.com/DeepanshS/mrsimulator-examples/
        master/spin_systems_v0.3.json>`_ of a JSON serialized file. For details, refer
        to the :ref:`load_spin_systems` section of this documentation.

        Args:
            str filename: A local or remote address to a JSON serialized file.

        Example
        -------

        >>> sim.load_spin_systems(filename) # doctest:+SKIP
        """
        contents = import_json(filename)
        # json_data = contents["spin_systems"]
        self.spin_systems = [SpinSystem.parse_dict_with_units(obj) for obj in contents]

    def export_spin_systems(self, filename: str):
        """
        Export a list of spin systems to a JSON serialized file.

        See an
        `example <https://raw.githubusercontent.com/DeepanshS/mrsimulator-examples/
        master/spin_systems_v0.3.json>`_ of a JSON serialized file. For details, refer
        to the :ref:`load_spin_systems` section.

        Args:
            str filename: A filename of the serialized file.

        Example
        -------

        >>> sim.export_spin_systems(filename) # doctest:+SKIP
        """
        spin_systems = [SpinSystem.to_dict_with_units(obj) for obj in self.spin_systems]
        with open(filename, "w", encoding="utf8") as outfile:
            json.dump(
                spin_systems,
                outfile,
                ensure_ascii=False,
                sort_keys=False,
                allow_nan=False,
            )

    def run(self, method_index=None, pack_as_csdm=True, **kwargs):
        """Run the simulation and compute spectrum.

        Args:
            method_index: An integer or a list of integers. If provided, only the
                simulations corresponding to the methods at the given index/indexes
                will be computed. The default is None, `i.e.`, the simulation for
                every method will be computed.
            bool pack_as_csdm: If true, the simulation results are stored as a
                `CSDM <https://csdmpy.readthedocs.io/en/stable/api/CSDM.html>`_ object,
                otherwise, as a `ndarray
                <https://numpy.org/doc/1.18/reference/generated/numpy.ndarray.html>`_
                object.
                The simulations are stored as the value of the
                :attr:`~mrsimulator.Method.simulation` attribute of the corresponding
                method.

        Example
        -------

        >>> sim.run() # doctest:+SKIP
        """

        if method_index is None:
            method_index = np.arange(len(self.methods))
        if isinstance(method_index, int):
            method_index = [method_index]
        for index in method_index:
            method = self.methods[index]
            amp, indexes = one_d_spectrum(
                method=method,
                spin_systems=self.spin_systems,
                **self.config.get_int_dict(),
                **kwargs,
            )

            self.indexes.append(indexes)

            if isinstance(amp, list):
                simulated_data = amp
            else:
                simulated_data = [amp]

            if pack_as_csdm:
                method.simulation = self._as_csdm_object(simulated_data, method)
            else:
                method.simulation = np.asarray(simulated_data)

    # """The frequency is in the units of Hz."""
    # gamma = method.isotope.gyromagnetic_ratio
    # B0 = method.spectral_dimensions[0].events[0].magnetic_flux_density
    # larmor_frequency = -gamma * B0
    # reference_offset_in_MHz =method.spectral_dimensions[0].reference_offset / 1e6
    # denom = reference_offset_in_MHz + larmor_frequency
    # freq = method.spectral_dimensions[0].coordinates_Hz / abs(denom)

    # freq *= u.Unit("ppm")
    # return freq, amp

    def save(self, filename: str, with_units=True):
        """Serialize the simulator object to a JSON file.

        Args:
            bool with_units: If true, the attribute values are serialized as physical
                quantities expressed as a string with a value and a unit. If false,
                the attribute values are serialized as floats.
            str filename: The filename of the serialized file.

        Example
        -------

        >>> sim.save('filename') # doctest: +SKIP
        """
        if not with_units:
            with open(filename, "w", encoding="utf8") as outfile:
                json.dump(
                    self.reduced_dict(),
                    outfile,
                    ensure_ascii=False,
                    sort_keys=False,
                    allow_nan=False,
                )
            return

        with open(filename, "w", encoding="utf8") as outfile:
            json.dump(
                self.to_dict_with_units(include_methods=True, include_version=True),
                outfile,
                ensure_ascii=False,
                sort_keys=False,
                allow_nan=False,
            )

    @classmethod
    def load(cls, filename: str, parse_units=True):
        """Load the :class:`~mrsimulator.Simulator` object from a JSON file by parsing.

        Args:
            bool parse_units: If true, parse the attribute values from the serialized
                file for physical quantities, expressed as a string with a value and a
                unit.
            str filename: The filename of a JSON serialized mrsimulator file.

        Returns:
            A :class:`~mrsimulator.Simulator` object.

        Example
        -------

        >>> sim_1 = sim.load('filename') # doctest: +SKIP

        .. seealso::
            :ref:`load_spin_systems`
        """
        contents = import_json(filename)

        if not parse_units:
            return Simulator(**contents)

        return Simulator.parse_dict_with_units(contents)

    def _as_csdm_object(self, data: np.ndarray, method: Method) -> cp.CSDM:
        """
        Converts the simulation data from the given method to a CSDM object. Read
        `csdmpy <https://csdmpy.readthedocs.io/en/latest/>`_ for details

        Return:
            A CSDM object.
        """
        new = cp.new()
        for dimension in method.spectral_dimensions:
            new.add_dimension(dimension.to_csdm_dimension())
            if new.dimensions[-1].origin_offset != 0:
                new.dimensions[-1].to("ppm", "nmr_frequency_ratio")

        dependent_variable = {
            "type": "internal",
            "quantity_type": "scalar",
            "numeric_type": "float64",
        }
        for index, datum in enumerate(data):
            if len(datum) == 0:
                continue

            dependent_variable["components"] = [datum]
            if self.config.decompose_spectrum == "spin_system":
                self._update_name_description_application(dependent_variable, index)

            new.add_dependent_variable(dependent_variable)
            new.dependent_variables[-1].encoding = "base64"
        return new

    def _update_name_description_application(self, obj, index):
        """Update the name and description of the dependent variable attributes
        using fields from the spin system."""
        label = self.spin_systems[index].label
        if label not in ["", None]:
            obj.update({"components_label": [label]})

        name = self.spin_systems[index].name
        name = name if name not in ["", None] else f"spin system {index}"
        obj.update({"name": name})

        description = self.spin_systems[index].description
        if description not in ["", None]:
            obj.update({"description": description})

        obj["application"] = {
            "com.github.DeepanshS.mrsimulator": {
                "spin_systems": [self.spin_systems[index].to_dict_with_units()]
            }
        }



# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import ClassVar
from typing import Dict
from typing import List

from mrsimulator.utils.parseable import Parseable
from pydantic import Field

from .transition_query import TransitionQuery

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class Event(Parseable):
    r"""Base Event class defines the spin environment and the transition query for a
    segment of the transition pathway.

    Attributes:
        fraction: A `required` float containing the weight of the frequency
            contribution from the event.
        magnetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field
            during the event in units of T. The default value is ``9.4``.
        rotor_frequency: An `optional` float containing the sample spinning frequency
            :math:`\nu_r`, during the event in units of Hz.
            The default value is ``0``.
        rotor_angle: An `optional` float containing the angle between the
            sample rotation axis and the applied external magnetic field,
            :math:`\theta`, during the event in units of rad.
            The default value is ``0.9553166``, i.e. the magic angle.
        transition_query: An `optional` TransitionQuery object or an equivalent dict
            object listing the queries used in selecting the active transitions
            during the event. Only the active transitions from this query
            contribute to the frequency.
    """

    fraction: float = 1.0
    magnetic_flux_density: float = Field(default=9.4, ge=0)
    rotor_frequency: float = Field(default=0.0, ge=0)
    # 54.735 degrees = 0.9553166 radians
    rotor_angle: float = Field(default=0.9553166, ge=0, le=1.5707963268)
    transition_query: TransitionQuery = TransitionQuery()
    user_variables: List = None

    property_unit_types: ClassVar = {
        "magnetic_flux_density": "magnetic flux density",
        "rotor_frequency": "frequency",
        "rotor_angle": "angle",
    }

    property_default_units: ClassVar = {
        "magnetic_flux_density": "T",
        "rotor_frequency": "Hz",
        "rotor_angle": "rad",
    }

    property_units: Dict = {
        "magnetic_flux_density": "T",
        "rotor_frequency": "Hz",
        "rotor_angle": "rad",
    }

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """
        Parse the physical quantities of an Event object from a python dictionary
        object.

        Args:
            dict py_dict: Dict object
        """
        py_dict_copy = deepcopy(py_dict)
        return super().parse_dict_with_units(py_dict_copy)



# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Union

import csdmpy as cp
import numpy as np
from mrsimulator.spin_system.isotope import Isotope
from mrsimulator.transition import Transition
from mrsimulator.transition.transition_list import TransitionList
from mrsimulator.utils.parseable import Parseable
from pydantic import validator

from .spectral_dimension import SpectralDimension
from .utils import cartesian_product
from .utils import D_symmetry_indexes
from .utils import P_symmetry_indexes
from .utils import query_permutations

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class Method(Parseable):
    r"""Base Method class. A method class represents the NMR method.

    Attributes
    ----------

    channels: list (optional).
        The value is a list of isotope symbols over which the given method applies.
        An isotope symbol is given as a string with the atomic number followed by its
        atomic symbol, for example, '1H', '13C', and '33S'. The default is an empty
        list.
        The number of isotopes in a `channel` depends on the method. For example, a
        `BlochDecaySpectrum` method is a single channel method, in which case, the
        value of this attribute is a list with a single isotope symbol, ['13C'].

        Example
        -------

        >>> bloch = Method()
        >>> bloch.channels = ['1H']

    spectral_dimensions: list of :ref:`spectral_dim_api` or dict objects (optional).
        The number of spectral dimensions depends on the given method. For example, a
        `BlochDecaySpectrum` method is a one-dimensional method and thus requires a
        single spectral dimension. The default is a single default
        :ref:`spectral_dim_api` object.

        Example
        -------

        >>> bloch = Method()
        >>> bloch.spectral_dimensions = [SpectralDimension(count=8, spectral_width=50)]
        >>> # or equivalently
        >>> bloch.spectral_dimensions = [{'count': 8, 'spectral_width': 50}]

    simulation: CSDM or ndarray (N/A).
        An object holding the result of the simulation. The initial value of this
        attribute is None. A value is assigned to this attribute when you run the
        simulation using the :meth:`~mrsimulator.Simulator.run` method.

    experiment: CSDM or ndarray (optional).
        An object holding the experimental measurement for the given method, if
        available. The default value is None.

        Example
        -------

        >>> bloch.experiment = my_data # doctest: +SKIP

    name: str (optional).
        The value is the name or id of the method. The default value is None.

        Example
        -------

        >>> bloch.name = 'BlochDecaySpectrum'
        >>> bloch.name
        'BlochDecaySpectrum'

    label: str (optional).
        The value is a label for the method. The default value is None.

        Example
        -------

        >>> bloch.label = 'One pulse acquired spectrum'
        >>> bloch.label
        'One pulse acquired spectrum'

    description: str (optional).
        The value is a description of the method. The default value is None.

        Example
        -------

        >>> bloch.description = 'Huh!'
        >>> bloch.description
        'Huh!'

    """
    name: str = None
    label: str = None
    description: str = None
    channels: List[str] = []
    spectral_dimensions: List[SpectralDimension] = [{}]
    simulation: Union[cp.CSDM, np.ndarray] = None
    experiment: Union[cp.CSDM, np.ndarray] = None

    property_default_units: ClassVar = {
        "magnetic_flux_density": "T",
        "rotor_angle": "rad",
        "rotor_frequency": "Hz",
    }

    property_units: Dict = {
        "magnetic_flux_density": "T",
        "rotor_angle": "rad",
        "rotor_frequency": "Hz",
    }

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @validator("channels", always=True)
    def validate_channels(cls, v, *, values, **kwargs):
        return [Isotope(symbol=_) for _ in v]

    @validator("experiment", pre=True, always=True)
    def validate_experiment(cls, v, *, values, **kwargs):
        if v is None:
            return None
        if isinstance(v, dict):
            return cp.parse_dict(v)
        if isinstance(v, cp.CSDM):
            return v
        raise ValueError("Unable to read the data.")

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse the physical quantity from a dictionary representation of the Method
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A python dict representation of the Method object.

        Returns:
            A :ref:`method_api` object.
        """
        py_dict_copy = deepcopy(py_dict)
        if "spectral_dimensions" in py_dict_copy:
            py_dict_copy["spectral_dimensions"] = [
                SpectralDimension.parse_dict_with_units(s)
                for s in py_dict_copy["spectral_dimensions"]
            ]
        if "simulation" in py_dict_copy:
            py_dict_copy["simulation"] = cp.parse_dict(py_dict_copy["simulation"])
        if "experiment" in py_dict_copy:
            py_dict_copy["experiment"] = cp.parse_dict(py_dict_copy["experiment"])
        return super().parse_dict_with_units(py_dict_copy)

    def update_spectral_dimension_attributes_from_experiment(self):
        """Update the spectral dimension attributes of the method to match the
        attributes of the experiment from the :attr:`~mrsimulator.Method.experiment`
        attribute."""
        spectral_dims = self.spectral_dimensions
        for i, dim in enumerate(self.experiment.dimensions):
            spectral_dims[i].count = dim.count
            spectral_dims[i].spectral_width = dim.count * dim.increment.to("Hz").value
            spectral_dims[i].reference_offset = dim.coordinates_offset.to("Hz").value
            spectral_dims[i].origin_offset = dim.origin_offset.to("Hz").value

    def to_dict_with_units(self):
        """
        Parse the class object to a JSON compliant python dictionary object where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit.

        Returns:
            A python dict object.
        """
        temp_dict = {}
        items = ["name", "label", "description"]
        for en in items:
            value = self.__getattribute__(en)
            if value is not None:
                temp_dict[en] = self.__getattribute__(en)

        temp_dict["spectral_dimensions"] = [
            item.to_dict_with_units() for item in self.spectral_dimensions
        ]
        temp_dict["channels"] = [item.to_dict_with_units() for item in self.channels]
        if self.simulation is not None:
            temp_dict["simulation"] = self.simulation.to_dict(update_timestamp=True)
        if self.experiment is not None:
            temp_dict["experiment"] = self.experiment.to_dict()
        return temp_dict

    def dict(self, **kwargs):
        temp_dict = super().dict(**kwargs)
        if self.simulation is not None:
            temp_dict["simulation"] = self.simulation.to_dict(update_timestamp=True)
        if self.experiment is not None:
            temp_dict["experiment"] = self.experiment.to_dict()
        return temp_dict

    def _get_transition_pathways(self, spin_system):
        all_transitions = spin_system._all_transitions()

        segments = []
        for seq in self.spectral_dimensions:
            selected_transitions = all_transitions[:]
            for ent in seq.events:
                list_of_P = query_permutations(
                    ent.transition_query.to_dict_with_units(),
                    isotope=spin_system.get_isotopes(),
                    channel=[item.symbol for item in self.channels],
                )
                indexes = P_symmetry_indexes(selected_transitions, list_of_P)
                selected_transitions = selected_transitions[indexes]

                if ent.transition_query.D is not None:
                    list_of_D = query_permutations(
                        ent.transition_query.to_dict_with_units(),
                        isotope=spin_system.get_isotopes(),
                        channel=[item.symbol for item in self.channels],
                        transition_symmetry="D",
                    )
                    indexes = D_symmetry_indexes(selected_transitions, list_of_D)
                    selected_transitions = selected_transitions[indexes]

                segments += [selected_transitions]
        return segments

    def get_transition_pathways(self, spin_system) -> np.ndarray:
        """
        Return a list of transition pathways from the given spin system that satisfy
        the query selection criterion of the method.

        Args:
            SpinSystem spin_system: A SpinSystem object.

        Returns:
            An array of TransitionList objects. Each TransitionList object is a
            transition pathways containing a series of Transition objects.
        """
        segments = self._get_transition_pathways(spin_system)
        segments = [
            np.asarray(
                TransitionList(
                    [
                        Transition(initial=_[0].tolist(), final=_[1].tolist())
                        for _ in item
                    ]
                )
            )
            for item in segments
        ]
        return cartesian_product(*segments)

    # def get_transition_pathways_old(self, spin_system):
    #     """
    #     Return a list of transition pathways from the given spin_system that satisfy
    #     the query criterion of the method.

    #     Args:
    #         spin_system: An SpinSystem object.
    #     """
    #     transitions = spin_system.all_transitions()
    #     segments = []
    #     for seq in self.spectral_dimensions:
    #         for ent in seq.events:
    #             list_of_P = query_permutations(
    #                 ent.transition_query.to_dict_with_units(),
    #                 isotope=spin_system.get_isotopes(),
    #                 channel=[item.symbol for item in self.channels],
    #             )
    #             P_segment = []
    #             # delta_P = transitions[:,:]
    #             # for symmetry in list_of_P:
    #             #     P_segment += transitions.filter(P=symmetry)

    #             # if ent.transition_query.D != None:
    #             #     list_of_D = query_permutations(
    #             #         ent.transition_query.to_dict_with_units(),
    #             #         isotope=spin_system.get_isotopes(),
    #             #         channel=[item.symbol for item in self.channels],
    #             #         transition_symmetry = "D"
    #             #     )

    #             #     D_segment = []
    #             #     for D_symmetry in list_of_D:
    #             #         D_segment += transitions.filter(D = D_symmetry)
    #             #     print('list of D: ', list_of_D)
    #             #     print('D_segment: ', D_segment)

    #             for symmetry in list_of_P:
    #                 if ent.transition_query.D is None:
    #                     P_segment += transitions.filter(P=symmetry)
    #                 elif ent.transition_query.D is not None:
    #                     list_of_D = query_permutations(
    #                         ent.transition_query.to_dict_with_units(),
    #                         isotope=spin_system.get_isotopes(),
    #                         channel=[item.symbol for item in self.channels],
    #                         transition_symmetry="D",
    #                     )
    #                     D_transition = []
    #                     [
    #                         D_transition.append(x)
    #                         for x in list_of_D
    #                         if x not in D_transition
    #                     ]
    #                     # D_segment = []
    #                     for D_symmetry in D_transition:
    #                         P_segment += transitions.filter(P=symmetry, D=D_symmetry)
    #                     # print('list of D: ', D_transition)
    #                     # for symmetry in list_of_D:
    #                     #     D_segment += transitions.filter(D=symmetry)
    #                     # print('D_segment: ', D_segment)

    #             segments.append(np.asarray(P_segment))  # append the intersection
    #     return cartesian_product(*segments)



# -*- coding: utf-8 -*-
class ZeemanState:
    """Zeeman energy state class."""

    def __init__(self, n_sites, *args):
        self.n_sites = n_sites
        for i in range(n_sites):
            self.__setattr__(f"m{i}", args[i])

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        lst = [int(2 * self.__getattribute__(f"m{i}")) for i in range(self.n_sites)]
        string = " ".join([rf"m_{i}={{{item} \over 2}}" for i, item in enumerate(lst)])
        return rf"$|{string}⟩$"

    def __str__(self):
        lst = ", ".join(
            [f"{self.__getattribute__(f'm{i}')}" for i in range(self.n_sites)]
        )
        return f"|{lst}⟩"

    def tolist(self):
        return [self.__getattribute__(f"m{i}") for i in range(self.n_sites)]



# -*- coding: utf-8 -*-
"""The AbstractOperation class."""
from mrsimulator.utils.parseable import Parseable


__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


class AbstractOperation(Parseable):
    """A base class for signal processing operations."""

    @property
    def function(self):
        return self.__class__.__name__

    def to_dict_with_units(self):
        my_dict = super().to_dict_with_units()
        my_dict["function"] = self.function
        if hasattr(self, "type"):
            my_dict["type"] = self.type
        return my_dict

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """Parse dictionary for SignalProcessor

        Args:
            py_dict (dict): A python dictionary representation of the operation.
        """
        my_dict_copy = py_dict.copy()
        if "function" in my_dict_copy.keys():
            my_dict_copy.pop("function")
        return super().parse_dict_with_units(my_dict_copy)



# -*- coding: utf-8 -*-
"""Base Parseable class."""
from copy import deepcopy
from typing import ClassVar
from typing import Dict

from csdmpy.units import string_to_quantity
from pydantic import BaseModel

from . import _reduce_dict

__author__ = "Shyam Dwaraknath"
__email__ = "shyamd@lbl.gov"


class Parseable(BaseModel):
    """
    Base class for all objects that can be parsed easily from JSON with units
    Don't directly use this. Rather inherit from it and implement a data model
    and property units and defaults
    """

    property_unit_types: ClassVar = {}

    property_default_units: ClassVar = {}

    property_units: Dict = {}

    @classmethod
    def parse_dict_with_units(cls, json_dict):
        """Parse the physical quantity from a dictionary representation of the class
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict json_dict: A required python dict object.
        """
        # Only consider properties with both unit types and default units
        all_props = set(cls.property_unit_types.keys()).intersection(
            set(cls.property_default_units.keys())
        )

        property_units = deepcopy(cls.property_default_units)
        for prop in all_props:
            required_type = cls.property_unit_types[prop]
            default_unit = cls.property_default_units[prop]

            if prop in json_dict:
                # If we have a single option
                if isinstance(required_type, str) and isinstance(default_unit, str):
                    try:
                        json_dict[prop] = enforce_units(
                            json_dict[prop], required_type, default_unit
                        )
                        property_units[prop] = default_unit
                    except Exception as e:
                        raise Exception(
                            f"Error enforcing units for {prop}: {json_dict[prop]}\n{e}"
                        )

                # If there are multiple type/unit combinations
                elif isinstance(required_type, list) and isinstance(default_unit, list):
                    pos_values = [
                        enforce_units(
                            json_dict[prop], r_type, d_unit, throw_error=False
                        )
                        for r_type, d_unit in zip(required_type, default_unit)
                    ]
                    # If none of the units were enforceable, error
                    # else choose the first good one
                    if not ([val is not None for val in pos_values]):
                        raise Exception(f"Could not enforce any units on {prop}")
                    else:
                        json_dict[prop], property_units[prop] = [
                            d for d in zip(pos_values, default_unit) if d[0] is not None
                        ][0]
        for k, v in property_units.items():
            property_units[k] = v[0] if isinstance(v, list) else v
        return cls(**json_dict, property_units=property_units)

    def reduced_dict(self, exclude=["property_units"]) -> dict:
        """Returns a reduced dictionary representation of the class object by removing
        all key-value pair corresponding to keys listed in the `exclude` argument, and
        keys with value as None.

        Args:
            exclude: A list of keys to exclude from the dictionary.
        Return: A dict.
         """
        return _reduce_dict(self.dict(), exclude)

    def to_dict_with_units(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object where
        the attribute value with physical quantity is expressed as a string with a
        number and a unit."""

        temp_dict = {}
        for k, v in self.dict(exclude={"property_units"}).items():

            # check the dict objects
            if isinstance(v, dict):
                val = getattr(self, k).to_dict_with_units()
                if val is not None:
                    temp_dict[k] = val

            # check the list objects
            elif isinstance(v, list):
                temp_dict[k] = get_list(getattr(self, k), v)
            elif v not in [None, ""]:
                temp_dict[k] = v

        temp_keys = temp_dict.keys()
        for key, unit in getattr(self, "property_units").items():
            if key in temp_keys:
                u = unit if unit != "pct" else "%"
                temp_dict[key] = f"{temp_dict[key]} {u}"
        return temp_dict


def enforce_units(value: str, required_type: str, default_unit: str, throw_error=True):
    """ Enforces a required type and default unit on the value. """
    try:
        value = string_to_quantity(value)
        data_type = value.unit.physical_type

        if required_type != data_type:
            raise Exception(
                f"A {required_type} value is required but got a {data_type} instead"
            )

        return value.to(default_unit).value
    except Exception as e:
        if throw_error:
            raise e
        else:
            return None


def get_list(member, obj):
    lst = []
    for i, item in enumerate(obj):
        if isinstance(item, list):
            lst.append(get_list(member[i], item))
        elif isinstance(item, dict):
            lst.append(member[i].to_dict_with_units())
        elif item not in [None, ""]:
            lst.append(item)
    return lst



# -*- coding: utf-8 -*-
"""Base ConfigSimulator class."""
from mrsimulator.sandbox import AveragingScheme
from pydantic import BaseModel
from pydantic import Field
from typing_extensions import Literal

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"

# decompose spectrum
__decompose_spectrum_enum__ = {"none": 0, "spin_system": 1}

# integration volume
__integration_volume_enum__ = {"octant": 0, "hemisphere": 1}
__integration_volume_octants__ = [1, 4]


class ConfigSimulator(BaseModel):
    r"""
    The configurable attributes for the Simulator class used in simulation.

    Attributes
    ----------

    number_of_sidebands: int (optional).
        The value is the requested number of sidebands that will be computed in the
        simulation. The value cannot be zero or negative. The default value
        is 64.

    integration_volume: enum (optional).
        The value is the volume over which the solid-state spectral frequency
        integration is performed. The valid literals of this enumeration are

        - ``octant`` (default), and
        - ``hemisphere``

    integration_density: int (optional).
        The value represents the integration density or equivalently the number of
        orientations over which the frequency integration is performed within a given
        volume. If :math:`n` is the integration_density, then the total number of
        orientation is given as

        .. math::
            n_\text{octants} \frac{(n+1)(n+2)}{2},

        where :math:`n_\text{octants}` is the number of octants in the given volume.
        The default value is 70.

    decompose_spectrum: enum (optional).
        The value specifies how a simulation result is decomposed into an array of
        spectra. The valid literals of this enumeration are

        - ``none`` (default): When the value is `none`, the resulting simulation is a
          single spectrum, which is an integration of the spectra over all spin systems.
        - ``spin_system``:  When the value is `spin_system`, the resulting simulation
          is an array of spectra, where each spectrum arises from a spin system within
          the Simulator object.

    Example
    -------

    >>> a = Simulator()
    >>> a.config.number_of_sidebands = 128
    >>> a.config.integration_density = 96
    >>> a.config.integration_volume = 'hemisphere'
    >>> a.config.decompose_spectrum = 'spin_system'
    """

    number_of_sidebands: int = Field(default=64, gt=0)
    integration_volume: Literal["octant", "hemisphere"] = "octant"
    integration_density: int = Field(default=70, gt=0)
    decompose_spectrum: Literal["none", "spin_system"] = "none"

    class Config:
        validate_assignment = True

    def get_int_dict(self):
        py_dict = self.dict()
        py_dict["integration_volume"] = __integration_volume_enum__[
            self.integration_volume
        ]
        py_dict["decompose_spectrum"] = __decompose_spectrum_enum__[
            self.decompose_spectrum
        ]
        return py_dict

    # averaging scheme. This contains the c pointer used in frequency evaluation
    # at c-level.
    @property
    def averaging_scheme(self):
        return self._averaging_scheme

    @averaging_scheme.setter
    def averaging_scheme(self, other):
        if isinstance(other, AveragingScheme):
            self._averaging_scheme = other
        raise ValueError("Expecting an instance of either the AveragingScheme class.")

    def get_orientations_count(self):
        """Return the total number of orientations.

        Example
        -------

        >>> a = Simulator()
        >>> a.config.integration_density = 20
        >>> a.config.integration_volume = 'hemisphere'
        >>> a.config.get_orientations_count() # (4 * 21 * 22 / 2) = 924
        924
        """
        n = self.integration_density
        vol = __integration_volume_octants__[
            __integration_volume_enum__[self.integration_volume]
        ]
        return int(vol * (n + 1) * (n + 2) / 2)



# -*- coding: utf-8 -*-
from functools import reduce
from itertools import permutations

import numpy as np

__author__ = ["Deepansh J. Srivastava", "Maxwell C. Venetos"]
__email__ = ["srivastava.89@osu.edu", "maxvenetos@gmail.com"]


def cartesian_product(*arrays):
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[..., i] = a
    return arr.reshape(-1, la)


def P_symmetry_indexes(transitions, list_of_P):
    P = transitions[:, 1, :] - transitions[:, 0, :]
    n = P.shape[1]
    return reduce(
        np.union1d,
        [
            reduce(np.intersect1d, [np.where(P[:, i] == search[i]) for i in range(n)])
            for search in list_of_P
        ],
        np.asarray([], dtype=np.int64),
    )


def D_symmetry_indexes(transitions, list_of_D):
    D = transitions[:, 1, :] ** 2 - transitions[:, 0, :] ** 2
    return reduce(
        np.union1d,
        [
            reduce(
                np.intersect1d,
                [np.where(D[:, i] == search[i]) for i in range(D.shape[1])],
            )
            for search in list_of_D
        ],
        np.asarray([], dtype=np.int64),
    )


def get_iso_dict(channel, isotope):
    """
        Parse the spin system sites to determine indices of each isotope that
        is part of the method channel.

        Args:
            channel: List object
            isotope: List object

    """
    iso_dict = {}

    # determine channels for P
    for i, item in enumerate(isotope):
        if item in channel and item not in iso_dict:
            iso_dict[item] = [i]
        elif item in iso_dict:
            iso_dict[item].append(i)

    return iso_dict


def query_permutations(query, isotope, channel, transition_symmetry="P"):
    """
        Determines the transition symmetries that are involved in a given
        transition query.

        Args:
            query: Dict object
            channel: List object
            isotope: List object
            transition_symmetry: str object. Derived from a transition query

    """

    P_permutated = []
    iso_dict = get_iso_dict(channel=channel, isotope=isotope)
    query_short = query[transition_symmetry]
    for i, items in enumerate(query_short):
        # Check if method isotope is in the spin system
        if channel[i] not in iso_dict:
            print(
                f"Method/channel isotope mismatch. Channel asks for {channel[i]} "
                f"but is not in {isotope}"
            )
            return []

        temp_P = []
        for k in range(len(query_short[items])):
            # Check transition query doesn't require more isotopes than present
            if len(query_short[items][k]) > len(iso_dict[channel[i]]):
                print("Failed: Transition query larger than channel")
                return []
            elif len(query_short[items][k]) <= len(iso_dict[channel[i]]):
                query_short[items][k] += (
                    len(iso_dict[channel[i]]) - len(query_short[items][k])
                ) * [k]
            temp_P += list(permutations(query_short[items][k]))
        P_permutated += [temp_P]
        # P_permutated += [list(permutations(query_short[items][k]))]

    transition_symmetry_from_query = []
    for i, iso_trans_symmetry in enumerate(P_permutated):
        # creating transition symmetries isotope by isotope
        temp_transitions = []
        for transition in iso_trans_symmetry:
            P_expanded = len(isotope) * [0]
            for j, item in enumerate(transition):
                # filling indices of spin system with a sites transition symmetries
                P_expanded[iso_dict[channel[i]][j]] = item

            if transition_symmetry_from_query == []:
                temp_transitions.append(P_expanded)
            else:
                # Each isotope is added to the previous isotope to create the
                # full transition symmetry
                for k, intermediate in enumerate(transition_symmetry_from_query[-1]):
                    temp_transitions.append(
                        [sum(x) for x in zip(intermediate, P_expanded)]
                    )

        transition_symmetry_from_query.append(temp_transitions)

    return transition_symmetry_from_query[-1]



cimport base_model as clib
from libcpp cimport bool as bool_t
from numpy cimport ndarray
import numpy as np
import cython
from mrsimulator import sandbox as sb

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


@cython.boundscheck(False)
@cython.wraparound(False)
def one_d_spectrum(method,
       list spin_systems,
       int verbose=0,
       int number_of_sidebands=90,
       unsigned int integration_density=72,
       unsigned int decompose_spectrum=0,
       unsigned int integration_volume=1,
       bool_t interpolation=True):
    """

    :ivar verbose:
        The allowed values are 0, 1, and 11. When the value is 1, the output is
        printed on the screen. When the value is 11, in addition to the output
        from 1, execution time is also printed on the screen.
        The default value is 0.
    :ivar number_of_sidebands:
        The value is an integer which corresponds to the number of sidebands
        simulated in the spectrum. The default value is 90. Note, when the
        sample spinning frequency is low, computation of more sidebands may be
        required for an acceptable result. The user is advised to ensure that
        enough sidebands are requested for computation.
    :ivar integration_density:
        The value is an integer which represents the frequency of class I
        geodesic polyhedra. These polyhedra are used in calculating the
        spherical average. Presently we only use octahedral as the frequency1
        polyhedra. As the frequency of the geodesic polyhedron increases, the
        polyhedra approach a sphere geometry. A higher frequency will result in a
        better powder averaging. The default value is 72.
        Read more on the `Geodesic polyhedron <https://en.wikipedia.org/wiki/Geodesic_polyhedron>`_.
    :ivar decompose_spectrum:
        An unsigned integer. When value is 0, the spectum is a sum of spectrum from all
        spin systems. If value is 1, spectrum from individual spin systems is stored
        separately.
    """

    cdef int transition_increment


# ---------------------------------------------------------------------
# observed spin _______________________________________________________
    # dimension = dimension[0]
    channel = method.channels[0].symbol
    # spin quantum number of the observed spin
    cdef double spin_quantum_number = method.channels[0].spin

    # gyromagnetic ratio
    # cdef double larmor_frequency =
    cdef gyromagnetic_ratio = method.channels[0].gyromagnetic_ratio
    cdef double factor = 1.0
    if gyromagnetic_ratio > 0.0:
        factor = -1.0

    # if verbose in [1, 11]:
    #     print(f'Simulating {isotope} (I={spin_quantum_number})')
    #     print(f'Larmor frequency (ω0 = - γ B0) = {larmor_frequency/1.0e6} MHz')
    #     print((f'Recording {isotope} spectrum with {number_of_points} '
    #             f'points over {spectral_width} Hz bandwidth'))
    #     print((f"and a reference offset of {dimension['reference_offset']} Hz."))

    # transitions of the observed spin
    cdef ndarray[float, ndim=1] transition_array
    cdef int number_of_transitions
    # transition_array = np.asarray([-0.5, 0.5]).ravel()
    # number_of_transitions = int(transition_array.size/2)
    # else:
    #     energy_level_count = int(2*spin_quantum_number+1)
    #     number_of_transitions = energy_level_count-1
    #     energy_states = np.arange(energy_level_count) - spin_quantum_number
    #     transitions = [ [energy_states[i], energy_states[i+1]] for i in range(number_of_transitions)]
    #     transition_array = np.asarray(transitions).ravel()


# -------------------------------------------------------------------------------------
# schemes

    cdef bool_t allow_fourth_rank = 0
    if spin_quantum_number > 0.5:
        allow_fourth_rank = 1

    # if sample_rotation_frequency_in_Hz < 1.0e-3:
    #     sample_rotation_frequency_in_Hz = 1.0e9
    #     rotor_angle_in_rad = 0.0
    #     number_of_sidebands = 1

# create averaging scheme _____________________________________________________
    cdef clib.MRS_averaging_scheme *the_averaging_scheme
    the_averaging_scheme = clib.MRS_create_averaging_scheme(
        integration_density=integration_density, allow_fourth_rank=allow_fourth_rank, integration_volume=integration_volume
    )

    max_n_sidebands = number_of_sidebands

# create sequences____________________________________________________________

    cdef int n_sequence = len(method.spectral_dimensions)
    total_n_points = 1
    cdef ndarray[int] n_event
    cdef ndarray[double] magnetic_flux_density_in_T
    cdef ndarray[double] srfiH
    cdef ndarray[double] rair
    cdef ndarray[int] cnt
    cdef ndarray[double] coord_off
    cdef ndarray[double] incre

    Bo = []
    vr = []
    th = []
    event_i = []
    count = []
    increment = []
    coordinates_offset = []

    prev_n_sidebands = 0
    for i, seq in enumerate(method.spectral_dimensions):
        for event in seq.events:
            if event.rotor_frequency < 1.0e-3:
                sample_rotation_frequency_in_Hz = 1.0e9
                rotor_angle_in_rad = 0.0
                number_of_sidebands = 1
                if prev_n_sidebands == 0: prev_n_sidebands = 1
            else:
                sample_rotation_frequency_in_Hz = event.rotor_frequency
                rotor_angle_in_rad = event.rotor_angle
                if prev_n_sidebands == 0: prev_n_sidebands = number_of_sidebands

            if prev_n_sidebands != number_of_sidebands:
                raise ValueError(
                    (
                        'The library does not support spectral dimensions containing '
                        'both zero and non-zero rotor frequencies. Consider using a '
                        'smaller value instead of zero.'
                    )
                )

            Bo.append(event.magnetic_flux_density)  # in T
            vr.append(sample_rotation_frequency_in_Hz) # in Hz
            th.append(rotor_angle_in_rad) # in rad

        total_n_points *= seq.count

        count.append(seq.count)
        offset = seq.spectral_width / 2.0
        coordinates_offset.append(-seq.reference_offset * factor - offset)
        increment.append(seq.spectral_width / seq.count)
        event_i.append(len(seq.events))

        seq.origin_offset = np.abs(Bo[0] * gyromagnetic_ratio * 1e6)

    magnetic_flux_density_in_T = np.asarray(Bo, dtype=np.float64)
    srfiH = np.asarray(vr, dtype=np.float64)
    rair = np.asarray(th, dtype=np.float64)
    cnt = np.asarray(count, dtype=np.int32)
    incre = np.asarray(increment, dtype=np.float64)
    coord_off = np.asarray(coordinates_offset, dtype=np.float64)
    n_event = np.asarray(event_i, dtype=np.int32)
    # magnetic_flux_density_in_T = np.asarray([dimension.magnetic_flux_density], dtype=np.float64)
    # srfiH = np.asarray([sample_rotation_frequency_in_Hz], dtype=np.float64)
    # rair = np.asarray([rotor_angle_in_rad], dtype=np.float64)
    # create spectral_dimensions



    the_sequence = clib.MRS_create_plans_for_sequence(
        the_averaging_scheme, &cnt[0], &coord_off[0],
        &incre[0], &magnetic_flux_density_in_T[0], &srfiH[0],
        &rair[0], &n_event[0], n_sequence, number_of_sidebands)

# normalization factor for the spectrum
    norm = np.prod(incre)

# create fftw scheme __________________________________________________________

    cdef clib.MRS_fftw_scheme *the_fftw_scheme
    the_fftw_scheme = clib.create_fftw_scheme(the_averaging_scheme.total_orientations, number_of_sidebands)
# # _____________________________________________________________________________


    # B0 = dimension.magnetic_flux_density

    # if verbose in [1, 11]:
    #     text = "`one_d_spectrum` method simulation parameters."
    #     len_ = len(text)
    #     print(text)
    #     print(f"{'-'*(len_-1)}")
    #     print (f'Macroscopic magnetic flux density (B0) = {B0} T')
    #     print (f'Sample rotation angle is (θ) = {rotor_angle_in_rad} rad')
    #     print (f'Sample rotation frequency (𝜈r) = {sample_rotation_frequency_in_Hz} Hz')

# sites _______________________________________________________________________________
    # CSA
    cdef int number_of_sites
    cdef ndarray[float] spin_i
    cdef ndarray[double] gyromagnetic_ratio_i

    cdef ndarray[double] iso_n
    cdef ndarray[double] zeta_n
    cdef ndarray[double] eta_n
    cdef ndarray[double] ori_n

    # quad
    cdef ndarray[double] Cq_e
    cdef ndarray[double] eta_e
    cdef ndarray[double] ori_e

    cdef ndarray[double] D_c

    cdef int trans__, pathway_increment, pathway_count, transition_count_per_pathway
    cdef ndarray[double, ndim=1] amp
    amp1 = np.zeros(total_n_points, dtype=np.float64)
    amp_individual = []

    cdef clib.isotopomer_ravel isotopomer_struct

    index_ = []
    # cdef clib.isotopomers_list *isotopomers_list_c

    # ---------------------------------------------------------------------
    # sample _______________________________________________________________
    for index, spin_sys in enumerate(spin_systems):
        abundance = spin_sys.abundance
        isotopes = [site.isotope.symbol for site in spin_sys.sites]
        if channel not in isotopes:
            continue

        # sub_sites = [site for site in spin_sys.sites if site.isotope.symbol == isotope]
        index_.append(index)
        number_of_sites= len(spin_sys.sites)

        if number_of_sites > 2:
            continue
        # site specification
        # CSA

        spin_i = np.empty(number_of_sites, dtype=np.float32)
        gyromagnetic_ratio_i = np.empty(number_of_sites, dtype=np.float64)

        iso_n = np.empty(number_of_sites, dtype=np.float64)
        zeta_n = np.empty(number_of_sites, dtype=np.float64)
        eta_n = np.empty(number_of_sites, dtype=np.float64)
        ori_n = np.zeros(3*number_of_sites, dtype=np.float64)

        # quad
        Cq_e = np.zeros(number_of_sites, dtype=np.float64)
        eta_e = np.zeros(number_of_sites, dtype=np.float64)
        ori_e = np.zeros(3*number_of_sites, dtype=np.float64)

        # for n sites, coupling grows as sum_{i=1}^{n-1}(i)
        D_c = np.zeros(number_of_sites, dtype=np.float64)

        # cdef int i, size = len(sample)

        amp = np.zeros(total_n_points)

        for i in range(number_of_sites):
            site = spin_sys.sites[i]
            spin_i[i] = site.isotope.spin
            gyromagnetic_ratio_i[i] = site.isotope.gyromagnetic_ratio


            # CSA tensor
            iso = site.isotropic_chemical_shift
            if iso is None: iso = 0.0

            shielding = site.shielding_symmetric
            if shielding is not None:
                zeta = shielding.zeta
                if zeta is None: zeta = 0.0
                eta = shielding.eta
                if eta is None: eta = 0.0
                alpha, beta, gamma = shielding.alpha, shielding.beta, shielding.gamma
                if alpha is None: alpha = 0.0
                if beta is None: beta = 0.0
                if gamma is None: gamma = 0.0
            else:
                zeta, eta, alpha, beta, gamma = 0.0, 0.0, 0.0, 0.0, 0.0

            # if verbose in [1, 11]:
            #     text = ((
            #         f"\n{isotope} site {i} from spin system {index_isotopomer} "
            #         f"@ {abundance}% abundance"
            #     ))
            #     len_ = len(text)
            #     print(text)
            #     print(f"{'-'*(len_-1)}")
            #     print(f'Isotropic chemical shift (δ) = {str(1e6*iso/larmor_frequency)} ppm')
            #     print(f'Shielding anisotropy (ζ) = {str(1e6*zeta/larmor_frequency)} ppm')
            #     print(f'Shielding asymmetry (η) = {eta}')
            #     print(f'Shielding orientation = [alpha = {alpha}, beta = {beta}, gamma = {gamma}]')

            # CSA tensor in Hz
            iso_n[i] = iso
            zeta_n[i] = zeta
            eta_n[i] = eta
            ori_n[3*i:3*i+3] = [alpha, beta, gamma]

            # quad tensor
            if spin_quantum_number > 0.5:
                quad = site.quadrupolar
                if quad is not None:
                    Cq = quad.Cq
                    if Cq is None: Cq = 0.0
                    eta = quad.eta
                    if eta is None: eta = 0.0
                    alpha, beta, gamma = quad.alpha, quad.beta, quad.gamma
                    if alpha is None: alpha = 0.0
                    if beta is None: beta = 0.0
                    if gamma is None: gamma = 0.0
                else:
                    Cq, eta, alpha, beta, gamma = 0.0, 0.0, 0.0, 0.0, 0.0

                Cq_e[i] = Cq
                eta_e[i] = eta
                ori_e[3*i:3*i+3] = [alpha, beta, gamma]

                # if verbose in [1, 11]:
                #     print(f'Quadrupolar coupling constant (Cq) = {Cq_e[i]/1e6} MHz')
                #     print(f'Quadrupolar asymmetry (η) = {eta}')
                #     print(f'Quadrupolar orientation = [alpha = {alpha}, beta = {beta}, gamma = {gamma}]')

        if number_of_sites != 0:
            transition_pathway = spin_sys.transition_pathways
            if transition_pathway is None:
                transition_pathway = method.get_transition_pathways(spin_sys)
            transition_pathway = np.asarray(transition_pathway)
            pathway_count, transition_count_per_pathway = transition_pathway.shape

            # convert transition objects to list
            lst = []
            for item in transition_pathway.ravel():
                lst += item.tolist()

            transition_array = np.asarray(lst, dtype=np.float32).ravel()
            pathway_increment = 2*number_of_sites*transition_count_per_pathway

            # if spin_sys.transitions is not None:
            #     transition_array = np.asarray(
            #         spin_sys.transitions, dtype=np.float32
            #     ).ravel()
            # else:
            #     transition_array = np.asarray([0.5, -0.5], dtype=np.float32)

            # the number 2 is because of single site transition [mi, mf]
            # it dose not work for coupled sites.
            # transition_increment = 2*number_of_sites
            # number_of_transitions = int((transition_array.size)/transition_increment)

            isotopomer_struct.number_of_sites = number_of_sites
            isotopomer_struct.spin = &spin_i[0]
            isotopomer_struct.gyromagnetic_ratio = &gyromagnetic_ratio_i[0]

            isotopomer_struct.isotropic_chemical_shift_in_ppm = &iso_n[0]
            isotopomer_struct.shielding_symmetric_zeta_in_ppm = &zeta_n[0]
            isotopomer_struct.shielding_symmetric_eta = &eta_n[0]
            isotopomer_struct.shielding_orientation = &ori_n[0]

            isotopomer_struct.quadrupolar_Cq_in_Hz = &Cq_e[0]
            isotopomer_struct.quadrupolar_eta = &eta_e[0]
            isotopomer_struct.quadrupolar_orientation = &ori_e[0]

            isotopomer_struct.dipolar_couplings = &D_c[0]

            # isotopomers_list_c[i] = isotopomer_struct
            for trans__ in range(pathway_count):
                clib.__mrsimulator_core(
                    # spectrum information and related amplitude
                    &amp[0],
                    &isotopomer_struct,
                    0, # turn off quad second order isotropic contribution
                    &transition_array[pathway_increment*trans__],
                    the_sequence,
                    n_sequence,
                    the_fftw_scheme,
                    the_averaging_scheme,
                    interpolation,
                    )

            temp = amp*abundance/norm

            ## reverse the spectrum if gyromagnetic ratio is positive.
            if gyromagnetic_ratio < 0:
                if total_n_points % 2 == 0:
                    temp[1:] = temp[1:][::-1]
                else:
                    temp = temp[::-1]

            if decompose_spectrum == 1:
                amp_individual.append(temp)
            else:
                amp1 += temp
        else:
            if decompose_spectrum == 1:
                amp_individual.append([])

    if decompose_spectrum == 1 and len(amp_individual) != 0:
        amp1 = amp_individual

    return amp1, index_



# -*- coding: utf-8 -*-
"""The Transition class."""
from typing import List

import numpy as np
from pydantic import BaseModel

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class Transition(BaseModel):
    """
    Base Transition class.

    Arguments:
        initial: The initial Zeeman energy state represented as a list of quantum
            numbers.
        final: The final Zeeman energy state represented as a list of quantum
            numbers.
    """

    initial: List[float] = []
    final: List[float] = []

    def __repr__(self):
        """Representation in bar-ket notation, |final⟩⟨initial|."""
        final = ", ".join([str(i) for i in self.final])
        initial = ", ".join([str(i) for i in self.initial])
        return f"|{final}⟩⟨{initial}|"

    # @property
    # def Zeeman_allowed(self):
    #     if abs(self.p) == 1:
    #         return True
    #     return False

    @property
    def p(self):
        """Return the total Δm (m_final-m_initial) value of the spin transition."""
        return self.P.sum()

    @property
    def delta_m(self):
        """An alias for p"""
        return self.p

    @property
    def P(self):
        """Return a list of Δm values of the spin transition for each site."""
        return np.asarray(self.final) - np.asarray(self.initial)

    @property
    def D(self):
        """Return a list of Δm**2 values of the spin transition for each site."""
        return np.asarray(self.final) ** 2 - np.asarray(self.initial) ** 2

    # def delta_m(self, i):
    #     """Return the Δm element of the transition corresponding to the ith site."""
    #     return self.final[i] - self.initial[i]

    def tolist(self):
        """Convert the transition to a list of quantum numbers where the first N
        quantum numbers corresponds to the initial energy state while the last N
        corresponds to the final energy state, where N is the number of sites."""
        lst = self.initial + self.final
        return lst

    def to_dict_with_units(self):
        return self.dict()



# -*- coding: utf-8 -*-
#
#  nmr_method.pxd
#
#  @copyright Deepansh J. Srivastava, 2019-2020.
#  Created by Deepansh J. Srivastava.
#  Contact email = deepansh2012@gmail.com
#
from libcpp cimport bool as bool_t

cdef extern from "angular_momentum.h":
    void wigner_d_matrices_from_exp_I_beta(int l, int n, void *exp_I_beta,
                                  double *wigner)

cdef extern from "schemes.h":
    ctypedef struct MRS_averaging_scheme:
        unsigned int total_orientations

    MRS_averaging_scheme *MRS_create_averaging_scheme(
                            unsigned int integration_density,
                            bool_t allow_fourth_rank,
                            unsigned int integration_volume)

    void MRS_free_averaging_scheme(MRS_averaging_scheme *scheme)

    ctypedef struct MRS_fftw_scheme:
        pass

    MRS_fftw_scheme *create_fftw_scheme(unsigned int total_orientations,
                                    int number_of_sidebands)

cdef extern from "mrsimulator.h":

    ctypedef struct MRS_plan:
        MRS_averaging_scheme *averaging_scheme
        int number_of_sidebands
        double sample_rotation_frequency_in_Hz
        double rotor_angle_in_rad
        # double complex *vector

    MRS_plan *MRS_create_plan(MRS_averaging_scheme *scheme, int number_of_sidebands,
                          double sample_rotation_frequency_in_Hz,
                          double rotor_angle_in_rad, double increment,
                          bool_t allow_fourth_rank)
    void MRS_free_plan(MRS_plan *plan)
    void MRS_get_amplitudes_from_plan(MRS_plan *plan, bool_t refresh)
    void MRS_get_frequencies_from_plan(MRS_plan *plan, double R0, double complex *R2,
                                  double complex *R4, bool_t refresh)

cdef extern from "isotopomer_ravel.h":
    ctypedef struct isotopomer_ravel:
        int number_of_sites                    # Number of sites
        float *spin                            # The spin quantum number
        double *gyromagnetic_ratio             # gyromagnetic ratio in (MHz/T)
        double *isotropic_chemical_shift_in_ppm # Isotropic chemical shift (Hz)
        double *shielding_symmetric_zeta_in_ppm     # Nuclear shielding anisotropy (Hz)
        double *shielding_symmetric_eta            # Nuclear shielding asymmetry
        double *shielding_orientation          # Nuclear shielding PAS to CRS euler angles (rad.)
        double *quadrupolar_Cq_in_Hz     # Quadrupolar coupling constant (Hz)
        double *quadrupolar_eta          # Quadrupolar asymmetry parameter
        double *quadrupolar_orientation        # Quadrupolar PAS to CRS euler angles (rad.)
        double *dipolar_couplings             # dipolar coupling stored as list of lists

cdef extern from "method.h":
    ctypedef struct MRS_event:
        double fraction                    # The weighted frequency contribution from the event.
        double magnetic_flux_density_in_T  #  he magnetic flux density in T.
        double rotor_angle_in_rad          # The rotor angle in radians.
        double sample_rotation_frequency_in_Hz # The sample rotation frequency in Hz.

    ctypedef struct MRS_sequence:
        int count                       #  The number of coordinates along the dimension.
        double increment                # Increment of coordinates along the dimension.
        double coordinates_offset       #  Start coordinate of the dimension.
        MRS_event *events               # Holds a list of events.
        unsigned int n_events           # The number of events.

    MRS_sequence *MRS_sequence_malloc(int n)

    MRS_sequence *MRS_create_plans_for_sequence(
        MRS_averaging_scheme *scheme,
        int *count,
        double *coordinates_offset,
        double *increment,
        double *magnetic_flux_density_in_T,
        double *sample_rotation_frequency_in_Hz,
        double *rotor_angle_in_rad,
        int *n_events,
        unsigned int n_seq,
        int number_of_sidebands)

cdef extern from "simulation.h":
    void mrsimulator_core(
        # spectrum information and related amplitude
        double * spec,
        double spectral_start,
        double spectral_increment,
        int number_of_points,

        isotopomer_ravel *ravel_isotopomer,
        MRS_sequence *the_sequence[],
        int n_sequence,

        int quad_second_order,                    # Quad theory for second order,
        bool_t remove_2nd_order_quad_isotropic,   # remove the isotropic contribution from the
                                                  # second order quad Hamiltonian.

        # spin rate, spin angle and number spinning sidebands
        int number_of_sidebands,
        double sample_rotation_frequency_in_Hz,
        double rotor_angle_in_rad,

        # The transition as transition[0] = mi and transition[1] = mf
        float *transition,
        int integration_density,
        unsigned int integration_volume,  # 0-octant, 1-hemisphere, 2-sphere
        bool_t interpolation
        )

    void __mrsimulator_core(
        # spectrum information and related amplitude
        double * spec,

        isotopomer_ravel *ravel_isotopomer,
        bool_t remove_2nd_order_quad_isotropic,   # remove the isotropic contribution from the
                                                  # second order quad Hamiltonian.

        # The transition as transition[0] = mi and transition[1] = mf
        float *transition,
        MRS_sequence *the_sequence, # the sequences within method.
        int n_sequence, # the number of sequences.
        MRS_fftw_scheme *fftw_scheme, # the fftw scheme
        MRS_averaging_scheme *scheme, # the powder averaging scheme
        bool_t interpolation
        )



# -*- coding: utf-8 -*-
"""Base Site class."""
from typing import ClassVar
from typing import Dict

from mrsimulator.utils.parseable import Parseable
from pydantic import validator

from .isotope import Isotope
from .tensors import AntisymmetricTensor
from .tensors import SymmetricTensor

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class Site(Parseable):
    """
    Base class representing a single-site nuclear spin interaction tensor parameters.
    The single-site nuclear spin interaction tensors include the nuclear shielding
    and the electric quadrupolar tensor.

    .. rubric:: Attribute Documentation

    Attributes
    ----------

    isotope: str (optional).
        A string expressed as an atomic number followed by an isotope symbol, eg.,
        `'13C'`, `'17O'`. The default value is `'1H'`.

        Example
        -------

        >>> site = Site(isotope='2H')

    isotropic_chemical_shift: float (optional).
        The value is the isotropic chemical shift of the site in the unit of ppm. The
        default value is 0.

        Example
        -------

        >>> site.isotropic_chemical_shift = 43.3

    shielding_symmetric: :ref:`sy_api` or equivalent dict object (optional).
        The value of this attribute represents the irreducible second-rank traceless
        symmetric part of the nuclear shielding tensor. The default value is None.

        The allowed attributes of the :ref:`sy_api` class for `shielding_symmetric` are
        ``zeta``, ``eta``, ``alpha``, ``beta``, and ``gamma``, where ``zeta`` is the
        shielding anisotropy, in ppm, and ``eta`` is the shielding asymmetry parameter
        defined using the Haeberlen convention. The Euler angles ``alpha``, ``beta``,
        and ``gamma`` are in radians.

        Example
        -------

        >>> site.shielding_symmetric = {'zeta': 10, 'eta': 0.5}
        >>> # or equivalently
        >>> site.shielding_symmetric = SymmetricTensor(zeta=10, eta=0.5)

    shielding_antisymmetric: :ref:`asy_api` or equivalent dict object (optional).
        The value of this attribute represents the irreducible first-rank antisymmetric
        part of the nuclear shielding tensor. The default value is None.

        The allowed attributes of the :ref:`asy_api` class for
        `shielding_antisymmetric` are ``zeta``, ``alpha``, and ``beta``, where ``zeta``
        is the anisotropy parameter of the anti-symmetric first-rank tensor given in
        ppm. The angles ``alpha`` and ``beta`` are in radians.

        Example
        -------

        >>> site.shielding_antisymmetric = {'zeta': 20}
        >>> # or equivalently
        >>> site.shielding_antisymmetric = AntisymmetricTensor(zeta=20)

    quadrupolar: :ref:`sy_api` or equivalent dict object (optional).
        The value of this attribute represents the irreducible second-rank traceless
        symmetric part of the electric-field gradient tensor. The default value is
        None.

        The allowed attributes of the :ref:`sy_api` class for `quadrupolar` are ``Cq``,
        ``eta``, ``alpha``, ``beta``, and ``gamma``, where ``Cq`` is the quadrupolar
        coupling constant, in Hz, and ``eta`` is the quadrupolar asymmetry parameter.
        The Euler angles ``alpha``, ``beta``, and ``gamma`` are in radians.

        Example
        -------

        >>> site.quadrupolar = {'Cq': 3.2e6, 'eta': 0.52}
        >>> # or equivalently
        >>> site.quadrupolar = SymmetricTensor(Cq=3.2e6, eta=0.52)

    name: str (optional).
        The value is the name or id of the site. The default value is None.

        Example
        -------

        >>> site.name = '2H-0'
        >>> site.name
        '2H-0'

    label: str (optional).
        The value is a label for the site. The default value is None.

        Example
        -------

        >>> site.label = 'Quad site'
        >>> site.label
        'Quad site'

    description: str (optional).
        The value is a description of the site. The default value is None.

        Example
        -------

        >>> site.description = 'An example Quadrupolar site.'
        >>> site.description
        'An example Quadrupolar site.'

    Example
    -------

    The following are a few examples of setting the site object.

    >>> site1 = Site(
    ...     isotope='33S',
    ...     isotropic_chemical_shift=20, # in ppm
    ...     shielding_symmetric={
    ...         "zeta": 10, # in ppm
    ...         "eta": 0.5
    ...     },
    ...     quadrupolar={
    ...         "Cq": 5.1e6, # in Hz
    ...         "eta": 0.5
    ...     }
    ... )

    Using SymmetricTensor objects.

    >>> site1 = Site(
    ...     isotope='13C',
    ...     isotropic_chemical_shift=20, # in ppm
    ...     shielding_symmetric=SymmetricTensor(zeta=10, eta=0.5),
    ... )
    """

    name: str = None
    label: str = None
    description: str = None
    isotope: str = "1H"
    isotropic_chemical_shift: float = 0
    shielding_symmetric: SymmetricTensor = None
    shielding_antisymmetric: AntisymmetricTensor = None
    quadrupolar: SymmetricTensor = None

    property_unit_types: ClassVar = {"isotropic_chemical_shift": "dimensionless"}
    property_default_units: ClassVar = {"isotropic_chemical_shift": "ppm"}
    property_units: Dict = {"isotropic_chemical_shift": "ppm"}

    @validator("quadrupolar")
    def spin_must_be_at_least_one(cls, v, values):
        if v is None:
            return v
        isotope = values["isotope"]
        isotope = Isotope(**isotope) if isinstance(isotope, dict) else isotope
        spin_I = isotope.spin
        if spin_I >= 1:
            if "zeta" in v.property_units:
                v.property_units.pop("zeta")
            return v
        message = (
            f"with spin quantum number {spin_I} does not allow quadrupolar tensor."
        )
        raise ValueError(f"{isotope} {message}")

    @validator("shielding_symmetric")
    def shielding_symmetric_must_not_contain_Cq(cls, v, values):
        if "Cq" in v.property_units:
            v.property_units.pop("Cq")
        return v

    @validator("isotope", always=True)
    def validate_isotope(cls, v, *, values, **kwargs):
        return Isotope(symbol=v)

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse the physical quantity from a dictionary representation of the Site
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A required python dict object.

        Returns:
            :ref:`site_api` object.

        Example
        -------

        >>> site_dict = {
        ...    "isotope": "13C",
        ...    "isotropic_chemical_shift": "20 ppm",
        ...    "shielding_symmetric": {"zeta": "10 ppm", "eta":0.5}
        ... }
        >>> site1 = Site.parse_dict_with_units(site_dict)
        """
        prop_mapping = {
            "shielding_symmetric": SymmetricTensor,
            "shielding_antisymmetric": AntisymmetricTensor,
            "quadrupolar": SymmetricTensor,
        }

        for k, v in prop_mapping.items():
            if k in py_dict:
                py_dict[k] = v.parse_dict_with_units(py_dict[k])

        return super().parse_dict_with_units(py_dict)

    def to_freq_dict(self, B0):
        """
        Serialize the Site object to a JSON compliant python dictionary object, where
        the attribute value is a number expressed in the attribute's default unit.
        The default unit for the attributes with respective dimensionalities is:

        - frequency: ``Hz``
        - angle: ``rad``

        Args:
            float B0: A required macroscopic magnetic flux density in units of T.

        Return:
            Python dict object.

        Example
        -------

        >>> pprint(site1.to_freq_dict(B0=9.4))
        {'description': None,
         'isotope': '13C',
         'isotropic_chemical_shift': -2013.1791999999998,
         'label': None,
         'name': None,
         'quadrupolar': None,
         'shielding_antisymmetric': None,
         'shielding_symmetric': {'alpha': None,
                                 'beta': None,
                                 'eta': 0.5,
                                 'gamma': None,
                                 'zeta': -1006.5895999999999}}
        """
        temp_dict = self.dict(exclude={"isotope"})
        temp_dict["isotope"] = self.isotope.symbol
        larmor_frequency = -self.isotope.gyromagnetic_ratio * B0  # in MHz
        for k in ["shielding_symmetric", "shielding_antisymmetric", "quadrupolar"]:
            if getattr(self, k):
                temp_dict[k] = getattr(self, k).to_freq_dict(larmor_frequency)
                if k == "shielding_symmetric":
                    temp_dict[k].pop("Cq")

        temp_dict["isotropic_chemical_shift"] *= larmor_frequency
        temp_dict.pop("property_units")
        return temp_dict



# -*- coding: utf-8 -*-
import platform
import sys
from os.path import abspath
from os.path import dirname
from os.path import join
from os.path import split

import numpy as np
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

try:
    from Cython.Build import cythonize

    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

# get the version from file
python_version = sys.version_info
py_version = ".".join([str(i) for i in python_version[:3]])
print("Using python version", py_version)
if python_version.major != 3 and python_version.minor < 6:
    print(
        f"Python version 3.6 and higher is required for the setup. You are using "
        f"version {py_version}"
    )
    sys.exit(1)

with open("src/mrsimulator/__init__.py", "r") as f:
    for line in f.readlines():
        if "__version__" in line:
            before_keyword, keyword, after_keyword = line.partition("=")
            version = after_keyword.strip()[1:-1]
            print("mrsimulator version ", version)

module_dir = dirname(abspath(__file__))


libraries = []
include_dirs = []
library_dirs = []
extra_compile_args = []
extra_link_args = []
data_files = []

numpy_include = np.get_include()

conda_location = numpy_include
for _ in range(5):
    conda_location = split(conda_location)[0]

if platform.system() == "Windows":
    # windows system lib and include path
    include_dirs += [join(conda_location, "Library", "include", "fftw")]
    include_dirs += [join(conda_location, "Library", "include", "openblas")]
    include_dirs += [join(conda_location, "Library", "include")]
    include_dirs += [join(conda_location, "include")]
    library_dirs += [join(conda_location, "Library", "lib")]
    extra_compile_args += ["-DFFTW_DLL"]

else:
    # unix system lib and include path
    conda_location = split(conda_location)[0]
    include_dirs += [join(conda_location, "include")]
    library_dirs += [join(conda_location, "lib")]
    extra_compile_args = ["-O3", "-ffast-math"]

libraries += ["fftw3", "openblas"]
extra_link_args += ["-lm"]

include_dirs = list(set(include_dirs))
library_dirs = list(set(library_dirs))
libraries = list(set(libraries))

# other include paths
include_dirs += ["src/c_lib/include/", numpy_include]

# print info
print(include_dirs)
print(library_dirs)
print(libraries)
print(extra_compile_args)
print(extra_link_args)

source = [
    "src/c_lib/lib/angular_momentum.c",
    "src/c_lib/lib/interpolation.c",
    "src/c_lib/lib/mrsimulator.c",
    "src/c_lib/lib/octahedron.c",
    "src/c_lib/lib/simulation.c",
    "src/c_lib/lib/powder_setup.c",
    "src/c_lib/lib/schemes.c",
    "src/c_lib/lib/method.c",
]

ext = ".pyx" if USE_CYTHON else ".c"

# method
ext_modules = [
    Extension(
        name="mrsimulator.base_model",
        sources=[*source, "src/c_lib/base/base_model" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        # data_files=data_files,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# tests
ext_modules += [
    Extension(
        name="mrsimulator.tests.tests",
        sources=[*source, "src/c_lib/test/test" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# sandbox
ext_modules += [
    Extension(
        name="mrsimulator.sandbox",
        sources=[*source, "src/c_lib/sandbox/sandbox" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

if USE_CYTHON:
    ext_modules = cythonize(ext_modules, language_level=3)

extras = {"lmfit": ["lmfit>=1.0.0"]}

description = "A python toolbox for simulating fast real-time solid-state NMR spectra."
setup(
    name="mrsimulator",
    version=version,
    description=description,
    long_description=open(join(module_dir, "README.md")).read(),
    author="Deepansh J. Srivastava",
    author_email="deepansh2012@gmail.com",
    python_requires=">=3.6",
    url="https://github.com/DeepanshS/MRsimulator/",
    packages=find_packages("src"),
    package_dir={"": "src"},
    extras_require=extras,
    ext_modules=ext_modules,
    include_package_data=True,
    zip_safe=False,
    license="BSD-3-Clause",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
)



# -*- coding: utf-8 -*-
from copy import deepcopy
from os import path

from monty.serialization import loadfn
from mrsimulator.method import Method
from mrsimulator.method.event import Event
from mrsimulator.method.spectral_dimension import SpectralDimension

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

# create class dynamically

MODULE_DIR = path.dirname(path.abspath(__file__))
METHODS_DATA = loadfn(path.join(MODULE_DIR, "methods_data.json"))

__doc_args__ = r"""

Args:
    spectral_dimensions: A list of python dict. Each dict is contains keywords that
        describe the coordinates along a spectral dimension. The keywords along with
        its definition are:

        count:
            An optional integer with the number of points, :math:`N`, along the
            dimension. The default value is 1024.
        spectral_width:
            An `optional` float with the spectral width, :math:`\Delta x`, along the
            dimension in units of Hz. The default is 25 kHz.
        reference_offset:
            An `optional` float with the reference offset, :math:`x_0` along the
            dimension in units of Hz. The default value is 0 Hz.
        origin_offset:
            An `optional` float with the origin offset (Larmor frequency) along the
            dimension in units of Hz. The default value is None.

    channels: A list of isotope symbols over which the method will be applied.
    rotor_frequency: An `optional` float containing the sample spinning frequency
        :math:`\nu_r`, in units of Hz. The default value is ``0``.
    rotor_angle: An `optional` float containing the angle between the sample rotation
        axis and the applied external magnetic field, :math:`\theta`, in units of rad.
        The default value is ``0.9553166``, i.e. the magic angle.
    magetic_flux_density: An `optional` float containing the macroscopic magnetic
        flux density, :math:`H_0`, of the applied external magnetic field in units of
        T. The default value is ``9.4``.

Return:
    A :class:`~mrsimulator.Method` instance.
"""


def prepare_method_structure(template, **kwargs):
    n_channels = template["number_of_channels"]
    prep = {"name": template["name"], "description": template["description"]}
    if "channels" in kwargs:
        prep["channels"] = kwargs["channels"]
        given_n_channels = len(prep["channels"])
        if given_n_channels != n_channels:
            raise ValueError(
                f"The method requires exactly {n_channels} channel(s), "
                f"{given_n_channels} provided."
            )
        kwargs.pop("channels")
    else:
        prep["channels"] = ["1H" for _ in range(n_channels)]
    return prep


def generate_method_from_template(template):
    # constructor
    def constructor(self, spectral_dimensions=[{}], parse=False, **kwargs):
        prep = prepare_method_structure(template, **kwargs)
        global_events = template["global_event_attributes"]
        ge = set(global_events)
        kw = set(kwargs)
        common = kw.intersection(ge)

        e = f"The attribute(s) {common} connot be modified for {prep['name']} method."
        if common != set():
            raise AttributeError(e)

        dim = []
        for i, s in enumerate(template["spectral_dimensions"]):
            events = []
            for e in s["events"]:
                ew = set(e)
                intersection = ew.intersection(kw)
                kw = kwargs.copy()
                for item in intersection:
                    kw.pop(item)
                if parse:
                    events.append({**e, **kw, **global_events})
                else:
                    events.append(Event(**e, **kw, **global_events))
            if parse:
                dim.append({**spectral_dimensions[i], "events": events})
            else:
                dim.append(SpectralDimension(**spectral_dimensions[i], events=events))

        if parse:
            method = {**prep, "spectral_dimensions": dim}
            return Method.parse_dict_with_units(method)
        else:
            return Method(**prep, spectral_dimensions=dim)

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse the physical quantities of the method object from as a python dictionary.

        Args:
            py_dict: Dict object.
        """
        dict_copy = deepcopy(py_dict)
        spectral_dimensions = dict_copy["spectral_dimensions"]
        dict_copy.pop("spectral_dimensions")
        return cls.__new__(
            0, spectral_dimensions=spectral_dimensions, parse=True, **dict_copy
        )

    description = template["description"]
    method = type(
        template["name"],
        (object,),
        {
            "__new__": constructor,
            "__str__": description,
            "__doc__": description + __doc_args__,
            "parse_dict_with_units": parse_dict_with_units,
        },
    )
    return method


# BlochDecaySpectrum
BlochDecaySpectrum = generate_method_from_template(METHODS_DATA["Bloch_decay"])

BlochDecayCentralTransitionSpectrum = generate_method_from_template(
    METHODS_DATA["Bloch_decay_central_transition"],
)



# -*- coding: utf-8 -*-
"""
Solid-state NMR spectra simulation module for Python
====================================================

mrsimulator is an incredibly fast solid-state NMR spectra simulation package,
capable of simulating spectra from both crystalline and amorphous-like materials.


It aims to provide simple and efficient solutions to the solid-state NMR spectrum
simulation problem. It includes tools for users to create their model spin systems,
simulate and compare the simulation with the measurement, and perform least-squares
minimization, using a collection of pre-defined NMR methods.

See https://mrsimulator.readthedocs.io/en/stable/ for complete documentation.
"""
# version has to be specified at the start.
__version__ = "0.3.0"

from .spin_system import Site  # lgtm [py/import-own-module]
from .spin_system import SpinSystem  # lgtm [py/import-own-module]
from .simulator import Simulator  # lgtm [py/import-own-module]
from .transition import Transition  # lgtm [py/import-own-module]
from .method.event import Event  # lgtm [py/import-own-module]
from .method.spectral_dimension import SpectralDimension  # lgtm [py/import-own-module]
from .method import Method  # lgtm [py/import-own-module]



# -*- coding: utf-8 -*-
import mrsimulator.signal_processing as sp
import mrsimulator.signal_processing.apodization as apo
from mrsimulator import Simulator

try:
    from lmfit import Parameters

    FOUND_LMFIT = True
except ImportError:
    FOUND_LMFIT = False


__author__ = "Maxwell C Venetos"
__email__ = "maxvenetos@gmail.com"

START = "sys_"
ENCODING_PAIRS = [
    ["spin_systems[", START],
    ["].sites[", "_site_"],
    ["].isotropic_chemical_shift", "_isotropic_chemical_shift"],
    ["].shielding_symmetric.", "_shielding_symmetric_"],
    ["].quadrupolar.", "_quadrupolar_"],
    ["].abundance", "_abundance"],
    ["methods[", "METHODS_"],  # why does methods need to be parameterized?
    ["].post_simulation", "_POST_SIM_"],
    [".scale", "scale"],
    [".apodization[", "APODIZATION_"],
    ["].args", "_args"],
]

EXCLUDE = [
    "property_units",
    "isotope",
    "name",
    "label",
    "description",
    "transition_pathways",
]


def _str_encode(my_string):
    """
    LMFIT Parameters class does not allow for names to include special characters.
    This function converts '[', ']', and '.' to their HTML numbers to comply with
    LMFIT.

    Args:
        my_string: A string object

    Returns:
        String object.
    """
    for item in ENCODING_PAIRS:
        my_string = my_string.replace(*item)
    return my_string


def _str_decode(my_string):
    """
    Converts the HTML numbers to '[', ']', and '.' to allow for execution of the
    parameter name to update the simulator.

    Args:
        my_string: A string object

    Returns:
        String Object.
    """
    for item in ENCODING_PAIRS:
        my_string = my_string.replace(*item[::-1])
    return my_string


def _list_of_dictionaries(my_list):
    """
    Helper function for traverse_dictionaries function which will return a list of
    dictionaries.

    Args:
        my_list: A list object

    Returns:
        List Object.
    """
    return [item.dict() for item in my_list]


def _traverse_dictionaries(dictionary, parent="spin_systems"):
    """
    Parses through the dictionary objects contained within the simulator object in
    order to return a list of all attributes that are populated.

    Args:
        dictionary: A dictionary or list object of the SpinSystem attributes from a
            simulation object
        parent: a string object used to create the addresses of the SpinSystem
            attributes.

    Returns:
        List Object.
    """
    name_list = []
    if isinstance(dictionary, dict):
        for key, vals in dictionary.items():
            if key not in EXCLUDE and vals is not None:
                if isinstance(vals, (dict, list)):
                    name_list += _traverse_dictionaries(
                        vals, _str_encode(f"{parent}.{key}")
                    )
                else:
                    name_list += [_str_encode(f"{parent}.{key}")]
    elif isinstance(dictionary, list):
        for i, items in enumerate(dictionary):
            name_list += _traverse_dictionaries(items, _str_encode(f"{parent}[{i}]"))

    return name_list


def _post_sim_LMFIT_params(post_sim):
    """
    Creates an LMFIT Parameters object for SignalProcessor operations
    involved in spectrum fitting

    Args:
        post_sim: SignalProcessor object

    Returns:
        Parameters object
    """
    temp_dict = {}
    # for item in post_sim.operations:
    #     prepend = f"DEP_VAR_{item.dependent_variable}_"
    for i, operation in enumerate(post_sim.operations):
        if isinstance(operation, apo.Gaussian):
            identifier = f"operation_{i}_Gaussian_FWHM"
            arg = operation.FWHM
            temp_dict[f"{identifier}"] = arg
        elif isinstance(operation, apo.Exponential):
            identifier = f"operation_{i}_Exponential_FWHM"
            arg = operation.FWHM
            temp_dict[f"{identifier}"] = arg
        elif isinstance(operation, sp.Scale):
            identifier = f"operation_{i}_Scale"
            arg = operation.factor
            temp_dict[f"{identifier}"] = arg

    params = Parameters()
    for key, val in temp_dict.items():
        params.add(name=key, value=val)

    return params


def _update_post_sim_from_LMFIT_params(params, post_sim):
    """Updates SignalProcessor operation arguments from an LMFIT Parameters object

    Args:
        params: LMFIT Parameters object
        post_sim: SignalProcessor object
    """
    temp_dict = {}
    arg_dict = {"Gaussian": "FWHM", "Exponential": "FWHM", "Scale": "factor"}
    for param in params:
        # iterating through the parameter list looking for only DEP_VAR
        # (ie post_sim params)
        if "operation_" in param:
            # splitting parameter name to obtain
            # Dependent variable index (var)
            # index of operation in the operation list (opIndex)
            # arg value for the operation (val)
            split_name = param.split("_")
            # var = split_name[split_name.index("VAR") + 1]
            opIndex = split_name[split_name.index("operation") + 1]
            val = params[param].value
            # creating a dictionary of operations and arguments for each dependent
            # variable
            # if f"DepVar_{var}" not in temp_dict.keys():
            #     temp_dict[f"DepVar_{var}"] = {}
            temp_dict[f"{opIndex}_{split_name[2]}"] = val

    # iterate through list of operation lists
    # for item in post_sim.operations:
    # iterating through dictionary with corresponding dependent variable index
    for operation, val in temp_dict.items():
        # creating assignment strings to create the correct address for updating each
        # operation
        split = operation.split("_")
        # dep_var_operation_list = f"post_sim.operations[{item.dependent_variable}]"
        operation_val_update = f"post_sim.operations[{split[0]}].{arg_dict[split[-1]]}"
        assignment = f"={val}"
        exec(operation_val_update + assignment)


def make_LMFIT_parameters(sim, post_sim=None, exclude_key=None):
    """
    Parses the Simulator and PostSimulator objects for a list of LMFIT parameters.
    The parameter name is generated using the following syntax:

    ``sys_i_site_j_attribute1_attribute2``

    for spin system attribute with signature sys[i].sites[j].attribute1.attribute2

    Args:
        sim: a Simulator object.
        post_sim: a SignalProcessor object

    Returns:
        LMFIT Parameters object.
    """
    if not FOUND_LMFIT:
        error = (
            f"The helper function {__name__} requires 'lmfit' module to create lmfit "
            r"paramters. Please install the lmfit module using\n'pip install lmfit'.",
        )
        raise ImportError(error)

    if not isinstance(sim, Simulator):
        raise ValueError(f"Expecting a `Simulator` object, found {type(sim).__name__}.")

    params = Parameters()
    temp_list = _traverse_dictionaries(_list_of_dictionaries(sim.spin_systems))

    # get total abundance scaling factor
    length = len(sim.spin_systems)
    abundance_scale = 100 / sum([sim.spin_systems[i].abundance for i in range(length)])

    # expression for the last abundance.
    last_abund = f"{length - 1}_abundance"
    expression = "-".join([f"{START}{i}_abundance" for i in range(length - 1)])
    expression = "100" if expression == "" else f"100-{expression}"
    for items in temp_list:
        if "_eta" in items:
            params.add(
                name=items, value=eval("sim." + _str_decode(items)), min=0, max=1,
            )
        # last_abund should come before abundance
        elif last_abund in items:
            params.add(
                name=items,
                value=eval("sim." + _str_decode(items)),
                min=0,
                max=100,
                expr=expression,
            )
        elif "abundance" in items:
            params.add(
                name=items,
                value=eval("sim." + _str_decode(items)) * abundance_scale,
                min=0,
                max=100,
            )
        else:
            value = eval("sim." + _str_decode(items))
            params.add(name=items, value=value)

    if post_sim is None:
        return params

    if not isinstance(post_sim, sp.SignalProcessor):
        raise ValueError(
            f"Expecting a `SignalProcessor` object, found {type(post_sim).__name__}."
        )
    temp_params = _post_sim_LMFIT_params(post_sim)
    for item in temp_params:
        params.add(name=item, value=temp_params[item].value)
    # params.add_many(temp_params)

    return params


def LMFIT_min_function(params, sim, post_sim=None):
    """
    The simulation routine to calculate the vector difference between simulation and
    experiment based on the parameters update.

    Args:
        params: Parameters object containing parameters to vary during minimization.
        sim: Simulator object used in the simulation. Initialized with guess fitting
            parameters.
        post_sim: PostSimulator object used in the simulation. Initialized with guess
            fitting parameters.

    Returns:
        Array of the differences between the simulation and the experimental data.
    """
    if not isinstance(params, Parameters):
        raise ValueError(
            f"Expecting a `Parameters` object, found {type(params).__name__}."
        )
    if not isinstance(post_sim, sp.SignalProcessor) or post_sim is None:
        raise ValueError(
            f"Expecting a `SignalProcessor` object, found {type(post_sim).__name__}."
        )

    if not isinstance(sim, Simulator):
        raise ValueError(f"Expecting a `Simulator` object, found {type(sim).__name__}.")

    values = params.valuesdict()
    for items in values:
        if "operation_" not in items:
            nameString = "sim." + _str_decode(items)
            executable = f"{nameString} = {values[items]}"
            exec(executable)
        elif "operation_" in items and post_sim is not None:
            _update_post_sim_from_LMFIT_params(params, post_sim)

    sim.run()
    processed_data = post_sim.apply_operations(data=sim.methods[0].simulation)
    # residual = np.asarray([])

    if sim.config.decompose_spectrum == "spin_system":
        datum = 0
        for decomposed_datum in processed_data.dependent_variables:
            datum += decomposed_datum.components[0]
            # datum = [sum(i) for i in zip(datum, decomposed_datum)]
    else:
        datum = processed_data.dependent_variables[0].components[0]

    return (
        sim.methods[0].experiment.dependent_variables[0].components[0].real - datum.real
    )

    # MULTIPLE EXPERIMENTS
    # for i, method in enumerate(sim.methods):
    #     y_factored = method.apodize().real
    #     residual = np.append(
    #         residual,
    #         method.experiment.dependent_variables[0].components[0].real - y_factored,
    #     )

    # return residual



# -*- coding: utf-8 -*-
import numpy as np
from mrsimulator.methods import BlochDecaySpectrum


def test_BlochDecaySpectrum():
    # test-1
    m1 = BlochDecaySpectrum()

    event_dictionary_ = {
        "fraction": 1.0,
        "magnetic_flux_density": "9.4 T",
        "rotor_frequency": "0.0 Hz",
        "rotor_angle": "0.9553166 rad",
        "transition_query": {"P": {"channel-1": [[-1.0]]}},
        "user_variables": ["magnetic_flux_density", "rotor_frequency", "rotor_angle"],
    }
    dimension_dictionary_ = {
        "count": 1024,
        "spectral_width": "25000.0 Hz",
        "reference_offset": "0.0 Hz",
        "events": [event_dictionary_],
    }

    should_be = {
        "name": "BlochDecaySpectrum",
        "description": "Simulate a 1D Bloch decay spectrum.",
        "channels": ["1H"],
        "spectral_dimensions": [dimension_dictionary_],
    }
    assert m1.to_dict_with_units() == should_be

    # test-2
    m2_dict = {
        "channels": ["29Si"],
        "magnetic_flux_density": "11.7 T",
        "rotor_angle": "90 deg",
        "spectral_dimensions": [{}],
    }
    m2 = BlochDecaySpectrum.parse_dict_with_units(m2_dict)

    angle = 90 * np.pi / 180
    event_dictionary_ = {
        "fraction": 1.0,
        "magnetic_flux_density": "11.7 T",
        "rotor_frequency": "0.0 Hz",
        "rotor_angle": f"{angle} rad",
        "transition_query": {"P": {"channel-1": [[-1.0]]}},
        "user_variables": ["magnetic_flux_density", "rotor_frequency", "rotor_angle"],
    }
    dimension_dictionary_ = {
        "count": 1024,
        "spectral_width": "25000.0 Hz",
        "reference_offset": "0.0 Hz",
        "events": [event_dictionary_],
    }

    should_be = {
        "name": "BlochDecaySpectrum",
        "description": "Simulate a 1D Bloch decay spectrum.",
        "channels": ["29Si"],
        "spectral_dimensions": [dimension_dictionary_],
    }

    assert m2.to_dict_with_units() == should_be



# -*- coding: utf-8 -*-
#
#  sandbox.pxd
#
#  @copyright Deepansh J. Srivastava, 2019-2020.
#  Created by Deepansh J. Srivastava
#  Contact email = deepansh2012@gmail.com
#

from libcpp cimport bool as bool_t

cdef extern from "schemes.h":
    ctypedef struct MRS_averaging_scheme:
        unsigned int total_orientations
        unsigned int integration_density
        unsigned int integration_volume

    MRS_averaging_scheme * MRS_create_averaging_scheme(
                            unsigned int integration_density,
                            bool_t allow_fourth_rank,
                            unsigned int integration_volume)

    MRS_averaging_scheme *MRS_create_averaging_scheme_from_alpha_beta(
                            double *alpha, double *beta,
                            double *weight, unsigned int n_angles,
                            bool_t allow_fourth_rank)

    void MRS_free_averaging_scheme(MRS_averaging_scheme *scheme)

cdef extern from "mrsimulator.h":

    ctypedef struct MRS_plan:
        MRS_averaging_scheme *averaging_scheme
        int number_of_sidebands
        double sample_rotation_frequency_in_Hz
        double rotor_angle_in_rad
        # double complex *vector

    MRS_plan *MRS_create_plan(MRS_averaging_scheme *scheme, int number_of_sidebands,
                          double sample_rotation_frequency_in_Hz,
                          double rotor_angle_in_rad, double increment,
                          bool_t allow_fourth_rank)
    void MRS_free_plan(MRS_plan *plan)
    void MRS_get_amplitudes_from_plan(MRS_plan *plan, double complex *R2,
                                  double complex *R4)
    void MRS_get_frequencies_from_plan(MRS_plan *plan, double R0)

    ctypedef struct MRS_dimension:
        pass



# -*- coding: utf-8 -*-
import platform
import sys
from os.path import abspath
from os.path import dirname
from os.path import exists
from os.path import join

import numpy as np
import numpy.distutils.system_info as sysinfo
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

from settings import use_accelerate
from settings import use_openblas

try:
    from Cython.Build import cythonize

    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

# get the version from file

python_version = sys.version_info
py_version = ".".join([str(i) for i in python_version[:3]])
print("Using python version", py_version)
if python_version.major != 3 and python_version.minor < 6:
    print(f"Python>=3.6 is required for the setup. You are using version {py_version}")
    sys.exit(1)

with open("src/mrsimulator/__init__.py", "r") as f:
    for line in f.readlines():
        if "__version__" in line:
            before_keyword, keyword, after_keyword = line.partition("=")
            version = after_keyword.strip()[1:-1]
            print("mrsimulator version ", version)

module_dir = dirname(abspath(__file__))


libraries = []
include_dirs = []
library_dirs = []
extra_compile_args = [
    "-O3",
    "-ffast-math",
    # "-msse4.2",
    # "-ftree-vectorize",
    # "-fopt-info-vec-optimized",
    # "-mavx",
]
extra_link_args = []
data_files = []

numpy_include = np.get_include()

if sys.platform.startswith("win"):

    extra_link_args += ["-Wl,--allow-multiple-definition"]
    extra_compile_args = ["-DFFTW_DLL"]

    # FFTW3 info
    fftw3_info = sysinfo.get_info("fftw3")
    fftw_keys = fftw3_info.keys()
    if "include_dirs" in fftw_keys:
        include_dirs += fftw3_info["include_dirs"]
    if "library_dirs" in fftw_keys:
        library_dirs += fftw3_info["library_dirs"]
    if "libraries" in fftw_keys:
        libraries += fftw3_info["libraries"]

    # OpenBLAS info
    openblas_info = sysinfo.get_info("openblas")
    if openblas_info != {}:
        include_dirs += [join(fftw3_info["include_dirs"][0], "openblas")]
        library_dirs += openblas_info["library_dirs"]
        libraries += openblas_info["libraries"]


def message(lib):
    return f"Please install {lib} from homebrew with:\n\t$ brew install {lib}"


if platform.system() == "Darwin":  # OSX-specific tweaks:
    # BLAS framework

    # Apple's Accelerate framework for BLAS:
    if use_accelerate:
        acc_info = sysinfo.get_info("accelerate")
        if "extra_compile_args" in acc_info:
            extra_compile_args += acc_info["extra_compile_args"]
        if "extra_link_args" in acc_info:
            extra_link_args += acc_info["extra_link_args"]

    # OpenBLAS framework
    if use_openblas:
        BLAS_INCLUDE = "/usr/local/opt/openblas/include"
        BLAS_LIB = "/usr/local/opt/openblas/lib"
        libraries += ["openblas"]

        if not exists(BLAS_INCLUDE):
            print(message("openblas"))
            sys.exit(1)

        include_dirs += [BLAS_INCLUDE]
        library_dirs += [BLAS_LIB]

    # # MKL framework
    # if use_mkl:
    #     mkl_info = np.__config__.blas_mkl_info
    #     if mkl_info == {}:
    #         print("Please enable mkl for numpy before proceeding.")
    #         sys.exit(1)

    #     BLAS_INCLUDE = mkl_info["include_dirs"]
    #     BLAS_LIB = mkl_info["library_dirs"]
    #     libraries += mkl_info["libraries"]

    #     include_dirs += BLAS_INCLUDE
    #     library_dirs += BLAS_LIB

    # FFTW framework
    FFTW_INCLUDE = "/usr/local/opt/fftw/include"
    FFTW_LIB = "/usr/local/opt/fftw/lib"
    libraries += ["fftw3"]

    if not exists(FFTW_INCLUDE):
        print(message("fftw"))
        sys.exit(1)

    include_dirs += [FFTW_INCLUDE]
    library_dirs += [FFTW_LIB]

    # if USE_SSE_AVX:
    #     extra_compile_args += ["-Wa,-q"]

if platform.system() == "Linux":
    include_dirs += [
        "/usr/include/",
        "/usr/include/openblas",
        "/usr/include/x86_64-linux-gnu/",
    ]

    library_dirs += ["/usr/lib64/", "/usr/lib/", "/usr/lib/x86_64-linux-gnu/"]
    libraries += ["openblas", "fftw3"]
    openblas_info = sysinfo.get_info("openblas")
    fftw3_info = sysinfo.get_info("fftw3")

    if openblas_info != {}:
        name = "openblas"
        library_dirs += openblas_info["library_dirs"]
        libraries += openblas_info["libraries"]
    fftw_keys = fftw3_info.keys()
    if "include_dirs" in fftw_keys:
        include_dirs += fftw3_info["include_dirs"]
    if "library_dirs" in fftw_keys:
        library_dirs += fftw3_info["library_dirs"]
    if "libraries" in fftw_keys:
        libraries += fftw3_info["libraries"]

    extra_link_args += ["-lm"]
    extra_compile_args += ["-g"]

include_dirs = list(set(include_dirs))
library_dirs = list(set(library_dirs))
libraries = list(set(libraries))

# other include paths
include_dirs += ["src/c_lib/include/", numpy_include]

# print info
print(include_dirs)
print(library_dirs)
print(libraries)
print(extra_compile_args)
print(extra_link_args)

source = [
    "src/c_lib/lib/angular_momentum.c",
    "src/c_lib/lib/interpolation.c",
    "src/c_lib/lib/mrsimulator.c",
    "src/c_lib/lib/octahedron.c",
    "src/c_lib/lib/simulation.c",
    "src/c_lib/lib/powder_setup.c",
    "src/c_lib/lib/schemes.c",
    "src/c_lib/lib/method.c",
]

ext = ".pyx" if USE_CYTHON else ".c"

# method
ext_modules = [
    Extension(
        name="mrsimulator.base_model",
        sources=[*source, "src/c_lib/base/base_model" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# tests
ext_modules += [
    Extension(
        name="mrsimulator.tests.tests",
        sources=[*source, "src/c_lib/test/test" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# sandbox
ext_modules += [
    Extension(
        name="mrsimulator.sandbox",
        sources=[*source, "src/c_lib/sandbox/sandbox" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

if USE_CYTHON:
    ext_modules = cythonize(ext_modules, language_level=3)

extras = {"lmfit": ["lmfit>=1.0.0"], "matplotlib": ["matplotlib>=3.0"]}

description = "A python toolbox for simulating fast real-time solid-state NMR spectra."
setup(
    name="mrsimulator",
    version=version,
    description=description,
    long_description=open(join(module_dir, "README.md")).read(),
    long_description_content_type="text/markdown",
    author="Deepansh J. Srivastava",
    author_email="deepansh2012@gmail.com",
    python_requires=">=3.6",
    url="https://github.com/DeepanshS/MRsimulator/",
    packages=find_packages("src"),
    package_dir={"": "src"},
    setup_requires=["numpy>=1.17"],
    install_requires=[
        "numpy>=1.17",
        "csdmpy>=0.3.2",
        "pydantic>=1.0",
        "monty>=2.0.4",
        "typing-extensions>=3.7",
    ],
    extras_require=extras,
    ext_modules=ext_modules,
    include_package_data=True,
    zip_safe=False,
    license="BSD-3-Clause",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
)



# -*- coding: utf-8 -*-
"""The Event class."""
from sys import modules
from typing import List
from typing import Union

import csdmpy as cp
from pydantic import BaseModel

from ._base import AbstractOperation

__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


class SignalProcessor(BaseModel):
    """
    Signal processing class to apply a series of operations to the dependent variables
    of the simulation dataset.

    Attributes
    ----------

    operations: List
        A list of operations.

    Examples
    --------

    >>> post_sim = SignalProcessor(operations=[o1, o2]) # doctest: +SKIP
    """

    processed_data: cp.CSDM = None
    operations: List[AbstractOperation] = []

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @classmethod
    def parse_dict_with_units(self, py_dict):
        """Parse a list of operations dictionary to a SignalProcessor class object.

        Args:
            pt_dict: A python dict object.
        """
        lst = []
        for op in py_dict["operations"]:
            if op["function"] == "apodization":
                lst.append(
                    getattr(
                        modules[__name__].apodization, op["type"]
                    ).parse_dict_with_units(op)
                )
            else:
                lst.append(
                    getattr(modules[__name__], op["function"]).parse_dict_with_units(op)
                )
        return SignalProcessor(operations=lst)

    def to_dict_with_units(self):
        """
        Serialize the SignalProcessor object to a JSON compliant python dictionary
        object, where physical quantities are represented as string with a value and a
        unit.

        Returns:
            A Dict object.
        """
        lst = []
        for i in self.operations:
            lst += [i.to_dict_with_units()]
        op = {}

        op["operations"] = lst
        return op

    def apply_operations(self, data, **kwargs):
        """
        Function to apply all the operation functions in the operations member of a
        SignalProcessor object. Operations applied sequentially over the data member.

        Returns:
            CSDM object: A copy of the data member with the operations applied to it.
        """
        if not isinstance(data, cp.CSDM):
            raise ValueError("The data must be a CSDM object.")
        copy_data = data.copy()
        for filters in self.operations:
            copy_data = filters.operate(copy_data)
        self.processed_data = copy_data

        return copy_data


class Scale(AbstractOperation):
    """
    Scale the amplitudes of all dependent variables from a CSDM object.

    Args:
        float factor: The scaling factor. The default value is 1.

    Example
    -------

    >>> import mrsimulator.signal_processing as sp
    >>> operation1 = sp.Scale(factor=20)
    """

    factor: float = 1

    def operate(self, data):
        r"""Applies the operation for which the class is named for.

        .. math::
            f(\vec(x)) = scale*\vec(x)

        Args:
            data: CSDM object
        """
        data *= self.factor
        return data


class IFFT(AbstractOperation):
    """
    Apply an inverse Fourier transform on all dependent variables of the CSDM object.

    Args:
        int dim_index: Dimension index along which the function is applied.

    Example
    -------

    >>> operation2 = sp.IFFT(dim_index=0)
    """

    dim_index: Union[int, list, tuple] = 0

    def operate(self, data):
        """Applies the operation for which the class is named for.

        Args:
            data: CSDM object
        """
        dim_index = self.dim_index
        if isinstance(dim_index, int):
            dim_index = [dim_index]

        for i in dim_index:
            data = data.fft(axis=i)
        return data


class FFT(IFFT):
    """
    Apply a forward Fourier transform on all dependent variables of the CSDM object.

    Args:
        int dim_index: Dimension index along which the function is applied.

    Example
    -------

    >>> operation3 = sp.FFT(dim_index=0)
    """

    pass


class complex_conjugate(AbstractOperation):
    pass



cimport sandbox as clib
from libcpp cimport bool as bool_t

cimport numpy as np
import numpy as np
import cython

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"

__integration_volume_enum__ = {"octant": 0, "hemisphere": 1}
__integration_volume_enum_rev__ = {0: "octant", 1: "hemisphere"}


# cdef class UserDefinedAveragingScheme:
#     cdef clib.MRS_averaging_scheme *scheme

#     def __init__(
#             self,
#             np.ndarray[ndim=1, double] alpha,
#             np.ndarray[ndim=1, double] beta,
#             np.ndarray[ndim=1, double] weight,
#             bool_t allow_fourth_rank=False
#         ):
#         """Create the generic averaging scheme for the simulation.""""
#         if alpha.size != beta.size != weight.size:
#             raise ValueError(
#                 'The length of alpha, beta, and weight array must be equal.'
#             )
#         cdef int n_angles = alpha.size
#         self.scheme = clib.MRS_create_averaging_scheme_from_alpha_beta(&alpha[0],
#                                             &beta[0], &weight[0], n_angles,
#                                             allow_fourth_rank_)
#     @property
#     def interpolation(self):
#         return False

#     @property
#     def total_orientations(self):
#         return self.scheme.total_orientations

#     def __reduce_cython__(self):
#         return 'AveragingScheme'

#     def __eq__(self, other):
#         if isinstance(other, AveragingScheme):
#             check = [
#                 self.alpha == other.alpha,
#                 self.beta == other.beta
#                 self.weight == other.weight
#             ]
#             if np.all(check):
#                 return True
#         return False

#     def __str__(self):
#         return self.__repr__()

#     def __repr__(self):
#         return (
#             "AveragingScheme(total_orientations={0}, interpolation={1})"
#         ).format(self.total_orientations, self.interpolation)

#     # def __del__(self):
#     #     clib.MRS_free_averaging_scheme(self.scheme)


cdef class AveragingScheme:
    cdef clib.MRS_averaging_scheme *scheme
    cdef bool_t allow_fourth_rank

    def __init__(self, int integration_density, integration_volume='octant', bool_t allow_fourth_rank=False):
        """Create the octahedral interpolation averaging scheme for the simulation.

        Args:
            integration_density: The number of triangles along the edge of the octahedron face.
            integration_volume: An enumeration literal, 'octant', 'hemisphere'.
            allow_fourth_rank: Boolean, If True, pre-calculates tables for computing fourth rank tensors.
        """
        self.allow_fourth_rank = allow_fourth_rank
        integration_volume_ = 0
        if integration_volume == 'hemisphere':
            integration_volume_=1
        self.scheme = clib.MRS_create_averaging_scheme(integration_density,
                                    allow_fourth_rank, integration_volume_)

    @property
    def interpolation(self):
        return True

    @property
    def total_orientations(self):
        return self.scheme.total_orientations

    # integration density
    @property
    def integration_density(self):
        return self.scheme.integration_density

    @integration_density.setter
    def integration_density(self, value):
        if isinstance(value, int):
            if value > 0:
                self.scheme = clib.MRS_create_averaging_scheme(value,
                                self.allow_fourth_rank,
                                self.scheme.integration_volume)
                return
        raise ValueError(f"Expecting a positive integer, found {value}.")

    # integration volume
    @property
    def integration_volume(self):
        return __integration_volume_enum_rev__[self.scheme.integration_volume]

    @integration_volume.setter
    def integration_volume(self, value):
        if value in __integration_volume_enum__.keys():
            self.scheme.integration_volume = __integration_volume_enum__[value]
            self.scheme = clib.MRS_create_averaging_scheme(
                    self.scheme.integration_density,
                    self.allow_fourth_rank, value
                )
            return
        raise ValueError(
            (
                "value is not a valid enumeration literal; "
                "permitted: 'octant', 'hemisphere', found {value}.",
            )
        )

    def __reduce_cython__(self):
        return 'AveragingScheme'

    def __eq__(self, other):
        if isinstance(other, AveragingScheme):
            check = [
                self.integration_volume == other.integration_volume,
                self.integration_density == other.integration_density
            ]
            if np.all(check):
                return True
        return False

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return (
            "AveragingScheme(total_orientations={0}, integration_density"
            "={1}, integration_volume={2}, interpolation={3})"
        ).format(
            self.total_orientations,
            self.integration_density,
            self.integration_volume,
            self.interpolation
        )

    def __del__(self):
        clib.MRS_free_averaging_scheme(self.scheme)


cdef class MRSPlan:
    cdef clib.MRS_plan *plan

    def __init__(self, AveragingScheme averaging_scheme, number_of_sidebands,
                sample_rotation_frequency_in_Hz, rotor_angle_in_rad,
                increment, allow_fourth_rank):

        self.plan = clib.MRS_create_plan(averaging_scheme.scheme,
                        number_of_sidebands, sample_rotation_frequency_in_Hz,
                        rotor_angle_in_rad, increment, allow_fourth_rank)

    @property
    def number_of_sidebands(self):
        return self.plan.number_of_sidebands

    @property
    def sample_rotation_frequency_in_Hz(self):
        return self.plan.sample_rotation_frequency_in_Hz

    @property
    def rotor_angle_in_rad(self):
        return self.plan.rotor_angle_in_rad

    # @property
    # def increment(self):
    #     return self.plan.increment

    # @property
    # def allow_fourth_rank(self):
    #     return self.plan.allow_fourth_rank

    # def evaluate(self, R0, R2, R4):
    #     cdef np.ndarray[double complex] R2_c = np.asarray(R2, dtype=np.complex128)
    #     cdef np.ndarray[double complex] R4_c = np.asarray(R4, dtype=np.complex128)
    #     cdef np.ndarray[double complex] output
    #     clib.MRS_get_amplitudes_from_plan(self.plan, &R2_c[0], &R4_c[0])
    #     clib.MRS_get_frequencies_from_plan(self.plan, R0)
    #     # side_band = np.
    #     output = self.plan.vector
    #     return output[::2].reshape(self.plan.number_of_sidebands,
    #                                 self.plan.averaging_scheme.total_orientations)


# @cython.boundscheck(False)
# @cython.wraparound(False)
# def MRS_plan(int integration_density,
#         int number_of_sidebands)



# -*- coding: utf-8 -*-
from typing import Dict
from typing import List
from typing import Optional

from mrsimulator.transition import Transition
from pydantic import BaseModel
from pydantic import Field

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class TransitionQuery(BaseModel):
    """Base TransitionQuery class.

    Attributes:
        P: A list of p symmetry transition, where p = Δm and Δm is the difference
                between spin quantum numbers of the final and initial states.
    """

    P: Optional[Dict] = {"channel-1": [[-1.0]]}
    D: Optional[Dict] = Field(default=None)
    f: Optional[Dict] = Field(default=None)
    transitions: List[Transition] = None

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def to_dict_with_units(self):
        return {k: v for k, v in self.dict().items() if v is not None}



# -*- coding: utf-8 -*-
"""The Event class."""
from sys import modules
from typing import Dict
from typing import Union

import numpy as np
from csdmpy.units import string_to_quantity
from pydantic import validator

from ._base import AbstractOperation

__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


class AbstractApodization(AbstractOperation):
    dim_index: Union[int, list, tuple] = 0
    dv_index: Union[int, list, tuple] = None  # if none apply to all

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        obj = super().parse_dict_with_units(py_dict)
        return getattr(modules[__name__], py_dict["type"])(**obj.dict())

    @property
    def function(self):
        return "apodization"

    @property
    def type(self):
        """The type apodization function."""
        return self.__class__.__name__

    @staticmethod
    def _get_dv_indexes(indexes, n):
        """Return a list of dependent variable indexes.

        Args:
            indexes: An interger, list of integers, or None indicating the dv indexes.
            n: Total number of dependent variables in the CSDM object.
        """
        if indexes is None:
            return np.arange(n)
        if isinstance(indexes, int):
            return [indexes]
        if isinstance(indexes, (list, tuple)):
            return np.asarray(indexes)

    def _operate(self, data, fn, prop_name, prop_value):
        """A generic operation function.

        Args:
            data: A CSDM object.
            fn: The apodization function.
            prop_name: The argument name for the function fn.
            prop_value: The argument value for the function fn.
        """
        dims = data.dimensions
        ndim = len(dims)

        dim_index = self.dim_index
        if isinstance(dim_index, int):
            dim_index = [dim_index]

        for dim_index_ in dim_index:
            x = dims[dim_index_].coordinates
            unit = 1 / self.property_units[prop_name]
            x = x.to(unit)
            x_value = x.value

            apodization_vactor = _get_broadcast_shape(
                fn(x_value, prop_value), dim_index_, ndim
            )

            n = len(data.dependent_variables)
            dv_indexes = self._get_dv_indexes(self.dv_index, n=n)

            for i in dv_indexes:
                data.dependent_variables[i].components *= apodization_vactor
        return data


def _str_to_quantity(v, values):
    if isinstance(v, str):
        quantity = string_to_quantity(v)
        values["property_units"] = {"FWHM": quantity.unit}
        return quantity.value
    if isinstance(v, float):
        return v


def _get_broadcast_shape(array, dim, ndim):
    """Return the broadcast shape of a vector `array` at dimension `dim` for `ndim`
    total dimensions. """
    none = [None for _ in range(ndim + 1)]
    if isinstance(dim, int):
        dim = [dim]
    for dim_ in dim:
        none[-dim_ - 1] = slice(None, None, None)
    return array[tuple(none)]


class Gaussian(AbstractApodization):
    r"""Apodize a dependent variable of the CSDM object with a Gaussian function.

    The apodization function follows

    .. math::
        f(x) = e^{-2 \pi^2 \sigma^2  x^2},

    where :math:`x` are the coordinates of the dimension, and :math:`\sigma` is the
    standard deviation. The relationship between the standard deviation, :math:`\sigma`,
    and the full width at half maximum of the reciprocal domain Gaussian function
    follows

    .. math::
        \sigma = \frac{\text{FWHM}}{2\sqrt{2\ln 2}}.

    Args:
        str FWHM: The full width at half maximum, FWHM, of the reciprocal domain
            Gaussian function, given as a string with a value and a unit. The default
            value is 0.
        int dim_index: The index of the CSDM dimension along which the operation is
            applied. The default is the dimension at index 0.
        int dv_index: The index of the CSDM dependent variable where the operation is
            applied. If the value is None, the operation will be applied to every
            dependent variable.

    Example
    -------

    >>> import mrsimulator.signal_processing.apodization as apo
    >>> operation4 = apo.Gaussian(FWHM='143.4 Hz', dim_index=0, dv_index=0)
    """

    FWHM: Union[float, str] = 0
    property_units: Dict = {"FWHM": ""}

    @validator("FWHM")
    def str_to_quantity(cls, v, values):
        return _str_to_quantity(v, values)

    # class Config:
    #     validate_assignment = True

    @staticmethod
    def fn(x, arg):
        # arg is FWHM
        sigma = arg / 2.354820045030949
        return np.exp(-2 * ((x * sigma * np.pi) ** 2))

    def operate(self, data):
        """
        Applies the operation for which the class is named for.

        data: CSDM object
        dep_var: int. The index of the dependent variable to apply operation to.
        """

        return self._operate(data, fn=self.fn, prop_name="FWHM", prop_value=self.FWHM)


class Exponential(AbstractApodization):
    r"""Apodize a dependent variable of the CSDM object by an exponential function.

    The apodization function follows

    .. math::
        f(x) = e^{-\Gamma |x| \pi},

    where :math:`x` are the coordinates of the dimension, and :math:`\Gamma` is the
    width parameter. The relationship between the width parameter, :math:`\Gamma`, and
    the full width at half maximum for the reciprocal domain Lorentzian function
    follows

    .. math::
        \text{FWHM} = \Gamma.

    Args:
        str FWHM: The full width at half maximum, FWHM, of the reciprocal domain
            Lorentzian function given as a string with a value and a unit. The default
            value is 0.
        int dim_index: The index of the CSDM dimension along which the operation is
            applied. The default is the dimension at index 0.
        int dv_index: The index of the CSDM dependent variable where the operation is
            applied. If the value is None, the operation will be applied to every
            dependent variable.

    Example
    -------

    >>> operation5 = apo.Exponential(FWHM='143.4 m', dim_index=0, dv_index=0)
    """

    FWHM: Union[float, str] = 0
    property_units: Dict = {"FWHM": ""}

    @validator("FWHM")
    def str_to_quantity(cls, v, values):
        return _str_to_quantity(v, values)

    # class Config:
    #     validate_assignment = True

    @staticmethod
    def fn(x, arg):
        return np.exp(-arg * np.pi * np.abs(x))

    def operate(self, data):
        """
        Applies the operation for which the class is named for.

        data: CSDM object
        dep_var: int. The index of the dependent variable to apply operation to.
        """

        return self._operate(data, fn=self.fn, prop_name="FWHM", prop_value=self.FWHM)


# class ExponentialAbs(AbstractApodization):
#     r"""Apodize a dependent variable of the simulation data object by an exponential
#     function. The function follows

#     .. math::
#         f(x) = e^{-\Tau |x| \pi},

#     where :math:`x` are the coordinates of the data dimension and :math:`\Tau` is
#     the width parameter.

#     Args:
#         int dim_index: Data dimension index to apply the function along.
#         float FWHM: The full width at half maximum parameter, :math:`\Tau`.
#         int dv_index: Data dependent variable index to apply the function to. If
#             the type None, the operation will be applied to every dependent variable.

#     Example
#     -------

#     >>> operation5 = apo.Exponential(FWHM=143.4, dim_index=0, dv_index=0)
#     """

#     FWHM: float = 0

#     property_unit_types: ClassVar = {"FWHM": ["time", "frequency"]}
#     property_default_units: ClassVar = {"FWHM": ["s", "Hz"]}
#     property_units: Dict = {"FWHM": "Hz"}

#     @staticmethod
#     def fn(x, arg):
#         return np.exp(-arg * np.pi * np.abs(x))

#     def operate(self, data):
#         """
#         Applies the operation for which the class is named for.

#         data: CSDM object
#         dep_var: int. The index of the dependent variable to apply operation to
#         """

#         return self._operate(data, fn=self.fn, prop_name="FWHM", prop_value=self.FWHM)



# -*- coding: utf-8 -*-


def get_spectral_dimensions(csdm_object):
    """
    Extract the count, spectral_width, and reference_offset parameters, associated with
    the spectral dimensions of the method, from the CSDM dimension objects.

    Args:
        csdm_object: A CSDM object holding the measurement dataset.

    Returns:
        A list of dict objects, where each dict containts the count, spectral_width, and
        reference_offset.
    """
    result = []
    for dim in csdm_object.dimensions:
        count = dim.count
        increment = dim.increment.to("Hz").value
        coordinates_offset = dim.coordinates_offset.to("Hz").value
        spectral_width = count * increment

        dim_i = {}
        dim_i["count"] = dim.count
        dim_i["spectral_width"] = spectral_width
        if dim.complex_fft is True:
            dim_i["reference_offset"] = coordinates_offset
        else:
            dim_i["reference_offset"] = coordinates_offset + spectral_width / 2.0

        result.append(dim_i)

    return result


def _reduce_dict(dict_obj, exclude=["property_units"]):
    """Reduce the dict by removing all key-value pair corresponding to keys listed in
    the `exclude` argument and keys with value as None.

    Args:
        exclude: List of keys to exclude from the dict.
    Returns: A dict.
    """
    if isinstance(dict_obj, dict):
        obj = {}
        for k, v in dict_obj.items():
            if k in exclude or v is None:
                continue
            if k == "isotope":
                obj[k] = v["symbol"]
                continue
            if k == "channels":
                obj[k] = [item["symbol"] for item in v]
                continue
            elif isinstance(v, dict):
                obj[k] = _reduce_dict(v)
            elif isinstance(v, list):
                obj[k] = [_reduce_dict(_) for _ in v]
            else:
                obj[k] = v
        return obj

    # if isinstance(dict_obj, list):
    #     obj = []
    #     for v in dict_obj:
    #         if v is None:
    #             continue
    #         if isinstance(v, dict):
    #             obj.append(_reduce_dict(v))
    #         elif isinstance(v, list):
    #             obj.append([_reduce_dict(_) for _ in v])
    #         else:
    #             obj.append(v)
    #     return obj

    return dict_obj
